"""

BinPan own indicators and utils.

"""
import numpy as np
import pandas as pd
from handlers.time_helper import convert_milliseconds_to_time_zone_datetime
from handlers.time_helper import pandas_freq_tick_interval


##############
# INDICATORS #
##############

def ichimoku(data: pd.DataFrame,
             tenkan: int = 9,
             kijun: int = 26,
             chikou_span: int = 26,
             senkou_cloud_base: int = 52,
             suffix: str = ''
             ) -> pd.DataFrame:
    """
    The Ichimoku Cloud is a collection of technical indicators that show support and resistance levels, as well as momentum and trend
    direction. It does this by taking multiple averages and plotting them on a chart. It also uses these figures to compute a “cloud”
    that attempts to forecast where the price may find support or resistance in the future.

    https://school.stockcharts.com/doku.php?id=technical_indicators:ichimoku_cloud

    https://www.youtube.com/watch?v=mCri-FFvZjo&list=PLv-cA-4O3y97HAd9OCvVKSfvQ8kkAGKlf&index=7

    :param pd.DataFrame data: A BinPan Symbol dataframe.
    :param int tenkan: The short period. It's the half sum of max and min price in the window. Default: 9
    :param int kijun: The long period. It's the half sum of max and min price in the window. Default: 26
    :param int chikou_span: Close of the next 26 bars. Util when spotting what happened with other ichimoku lines and what happened
      before Default: 26.
    :param senkou_cloud_base: Period to obtain kumo cloud base line. Default is 52.
    :param str suffix: A decorative suffix for the name of the column created.
    :return pd.DataFrame: A pandas dataframe with columns as indicator lines.

       .. code-block:: python

            from binpan import binpan

            sym = binpan.Symbol(symbol='LUNCBUSD', tick_interval='1m', limit=500)
            sym.ichimoku()
            sym.plot()

       .. image:: images/indicators/ichimoku.png
        :width: 1000
        :alt: Candles with some indicators

    """
    df = data.copy(deep=True)

    tick_interval = df.index.name.split()[1]
    df = df.asfreq(pandas_freq_tick_interval[tick_interval])

    high = df['High']
    low = df['Low']
    close = df['Close']
    # freq = pd.infer_freq(df.index)

    tenkan_sen = (high.rolling(window=tenkan).max() + low.rolling(window=tenkan).min()) / 2
    kijun_sen = (high.rolling(window=kijun).max() + low.rolling(window=kijun).min()) / 2

    chikou_span_serie = close.shift(periods=-chikou_span, freq='infer')

    arr_a = (tenkan_sen + kijun_sen) / 2
    senkou_span_a = arr_a.shift(periods=chikou_span, freq='infer')

    arr_b = (high.rolling(window=senkou_cloud_base).max() + low.rolling(window=senkou_cloud_base).min()) / 2
    senkou_span_b = arr_b.shift(periods=chikou_span, freq='infer')

    ret = pd.DataFrame([tenkan_sen, kijun_sen, chikou_span_serie, senkou_span_a, senkou_span_b]).T

    if suffix:
        suffix = '_' + suffix

    ret.columns = [f"Ichimoku_tenkan_{tenkan}" + suffix,
                   f"Ichimoku_kijun_{kijun}" + suffix,
                   f"Ichimoku_chikou_span{chikou_span}" + suffix,
                   f"Ichimoku_cloud_{tenkan}_{kijun}" + suffix,
                   f"Ichimoku_cloud_{senkou_cloud_base}" + suffix]
    # return ret.dropna(how='all', inplace=True)
    return ret


def ker(close: pd.Series,
        window: int,
        ) -> pd.Series:
    """
    Kaufman's Efficiency Ratio based in: https://stackoverflow.com/questions/36980238/calculating-kaufmans-efficiency-ratio-in-python-with-pandas

    :param pd.Series close: Close prices serie.
    :param int window: Window to check indicator.
    :return pd.Series: Results.
    """
    direction = close.diff(window).abs()
    volatility = pd.rolling_sum(close.diff().abs(), window)
    return direction / volatility


def fractal_w(data: pd.DataFrame,
              period=2,
              merged: bool = True,
              suffix: str = '',
              ) -> pd.DataFrame:
    """
    The fractal indicator is based on a simple price pattern that is frequently seen in financial markets. Outside of trading, a fractal
    is a recurring geometric pattern that is repeated on all time frames. From this concept, the fractal indicator was devised.
    The indicator isolates potential turning points on a price chart. It then draws arrows to indicate the existence of a pattern.

    https://www.investopedia.com/terms/f/fractal.asp

    From: https://codereview.stackexchange.com/questions/259703/william-fractal-technical-indicator-implementation

    :param pd.DataFrame data: BinPan dataframe with High prices.
    :param int period: Default is 2. Count of neighbour candles to match max or min tags.
    :param bool merged: If True, values are merged into one pd.Serie. minimums overwrite maximums in case of coincidence.
    :param str suffix: A decorative suffix for the name of the column created.
    :return pd.Series: A serie with 1 or -1 for local max or local min to tag.
    """
    window = 2 * period + 1  # default 5

    mins = data['Low'].rolling(window, center=True).apply(lambda x: x[period] == min(x), raw=True)
    maxs = data['High'].rolling(window, center=True).apply(lambda x: x[period] == max(x), raw=True)

    mins = mins.replace({0: np.nan})
    maxs = maxs.replace({0: np.nan})
    maxs.loc[:] = maxs * -1

    # return mins, maxs
    if suffix:
        suffix = '_' + suffix

    mins.name = f"Fractal_W_Min_{period}" + suffix
    maxs.name = f"Fractal_W_Max_{period}" + suffix

    min_idx = mins.dropna().index
    max_idx = maxs.dropna().index

    values = pd.Series(np.nan, index=data.index)
    values.loc[min_idx] = data['Low']
    values.loc[max_idx] = data['High']
    values.name = f"Fractal_W_{period}_values" + suffix

    if not merged:
        return pd.DataFrame([mins, maxs, values]).T

    else:
        merged = mins
        merged.fillna(maxs, inplace=True)
        merged.name = f"Fractal_W_{period}" + suffix
        return pd.DataFrame([merged, values]).T


####################
# INDICATORS UTILS #
####################

def split_serie_by_position(serie: pd.Series,
                            splitter_serie: pd.Series,
                            fill_with_zeros: bool = True) -> pd.DataFrame:
    """
    Splits a serie by values of other serie in four series by relative positions for plotting colored clouds with plotly.

    This means you will get 4 series with different situations:

       - serie is over the splitter serie.
       - serie is below the splitter serie.
       - splitter serie is over the serie.
       - splitter serie is below the serie.

    :param pd.Series serie: A serie to classify in reference to other serie.
    :param pd.Series splitter_serie: A serie to split in two couple of series classified by position reference.
    :param bool fill_with_zeros: Fill nans with zeros for splitted lines like MACD to avoid artifacts in plots.
    :return tuple: A tuple with four series classified by upper position or lower position.
    """
    serie_up = serie.loc[serie.ge(splitter_serie)]
    # splitter_up = splitter_serie.loc[serie_up.index]
    splitter_up = splitter_serie.loc[serie.ge(splitter_serie)]

    serie_down = serie.loc[serie.lt(splitter_serie)]
    # splitter_down = splitter_serie.loc[serie_down.index]
    splitter_down = splitter_serie.loc[serie.lt(splitter_serie)]

    serie_up.name += '_up'
    serie_down.name += '_down'
    splitter_up.name += '_split_up'
    splitter_down.name += '_split_down'

    # return serie_up, serie_splitter_up, serie_down, serie_splitter_down
    # return serie_up, serie_down
    empty_df = pd.DataFrame(index=serie.index)

    empty_df.loc[serie_up.index, serie_up.name] = serie_up
    empty_df.loc[splitter_up.index, splitter_up.name] = splitter_up

    empty_df.loc[serie_down.index, serie_down.name] = serie_down
    empty_df.loc[splitter_down.index, splitter_down.name] = splitter_down

    # ret = pd.DataFrame([serie_up, splitter_up, serie_down, splitter_down]).T

    if fill_with_zeros:
        return empty_df.fillna(0)
    else:
        return empty_df


def df_splitter(data: pd.DataFrame, up_column: str, down_column: str) -> list:
    """
    Splits a dataframe y sub dataframes to plot by colored areas.

    Based on: https://stackoverflow.com/questions/64741015/plotly-how-to-color-the-fill-between-two-lines-based-on-a-condition

    :param pd.DataFrame data: Indicator Dataframe
    :param str up_column: Name of the column to plot green when up side.
    :param str down_column: Name of the column to plot green when down side.
    :return list: A list with splitted dataframes to plot for.
    """

    df = data.copy(deep=True)

    # split data into chunks where series cross each other
    df['label'] = np.where(df[up_column].gt(df[down_column]), 1, 0)
    df['group'] = df['label'].ne(df['label'].shift()).cumsum()
    df = df.groupby('group')
    dfs = []

    for name, data in df:
        dfs.append(data)
    return dfs


def zoom_cloud_indicators(plot_splitted_serie_couples: dict,
                          main_index: list,
                          start_idx: int,
                          end_idx: int) -> dict:
    """
    It zooms the cloud indicators in an index interval for a plot zoom.

    :param dict plot_splitted_serie_couples: Splitted indicators for cloud colored area plotting.
    :param list main_index: The BinPan general index for cutting.
    :param int start_idx: A index to cut.
    :param int end_idx: A index to cut.
    :return dict: All indicators cut.
    """
    try:
        assert start_idx < end_idx <= len(main_index)
    except AssertionError:
        raise Exception(f"BinPan Plot Error: Zoom index not valid. Not start={start_idx} < end={end_idx} < len={len(main_index)}")

    ret = {}
    my_start = main_index[start_idx]
    my_end = main_index[end_idx]
    for k, v in plot_splitted_serie_couples.items():
        splits = v[2]
        cut_splits = []
        for df in splits:
            result = df.loc[my_start:my_end]
            if not result.empty:
                cut_splits.append(result)

        ret[k] = [v[0], v[1], cut_splits, v[3], v[4]]
    return ret


def shift_indicator(serie: pd.Series, window: int = 1):
    """
    It shifts a candle ahead by the window argument value (or backwards if negative).

    Just works with time indexes.

    :param pd.Series serie: A pandas Series.
    :param int window: Times values are shifted ahead. Default is 1.
    :return pd.Series: A series with index adjusted to the new shifted positions of values.
    """
    return serie.shift(window, freq='infer')


def ffill_indicator(serie: pd.Series, window: int = 1):
    """
    It forward fills a value through nans while a window of candles ahead.

    :param pd.Series serie: A pandas Series.
    :param int window: Times values are shifted ahead. Default is 1.
    :return pd.Series: A series with index adjusted to the new shifted positions of values.
    """
    return serie.ffill(limit=window)


###############
# From trades #
###############

def reversal_candles(trades: pd.DataFrame,
                     decimal_positions: int,
                     time_zone: str,
                     min_height: int = 7,
                     min_reversal: int = 4) -> pd.DataFrame:
    """
    Generate reversal candles for reversal charts:
       https://atas.net/atas-possibilities/charts/how-to-set-reversal-charts-for-finding-the-market-reversal/

    :param pd.Series trades: A dataframe with trades sizes, side and prices.
    :param int decimal_positions: Because this function uses integer numbers for prices, is needed to convert prices. Just steps are relevant.
    :param str time_zone: A time zone like "Europe/Madrid".
    :param int min_height: Minimum candles height in pips.
    :param int min_reversal: Maximum reversal to close the candles
    :return pd.DataFrame: A serie with the resulting candles number sequence.

     Example:
        .. code-block:: python

           from binpan import binpan

            ltc = binpan.Symbol(symbol='ltcbtc',
                                tick_interval='5m',
                                time_zone = 'Europe/Madrid',
                                time_index = True,
                                closed = True,
                                hours=5)
           ltc.get_trades()
           ltc.get_reversal_candles()
           ltc.plot_reversal()

        .. image:: images/indicators/reversal.png
           :width: 1000

    """

    prices = (trades['Price'].to_numpy() * 10 ** decimal_positions).astype(int)
    height = 0

    candle = 0
    prev_candle = 0
    candle_ids = [0]  # first value in a candle

    current_open = prices[0]
    open_ = [current_open]
    high = [current_open]
    low = [current_open]
    close = [current_open]
    prices_pool = [current_open]
    # reversal_control = [0]

    for idx in range(1, prices.size):
        current_price = prices[idx]
        previous_price = prices[idx - 1]
        delta = current_price - previous_price

        if candle != prev_candle:
            current_open = current_price

        prices_pool.append(current_price)
        current_high = max(prices_pool)
        current_low = min(prices_pool)

        # collect results
        open_.append(current_open)
        high.append(current_high)
        low.append(current_low)
        close.append(current_price)
        candle_ids.append(candle)

        # update height if not accomplished, else is fixed
        if np.abs(height) < min_height:
            height += delta

        height_bool = np.abs(height) >= min_height
        bull_reversal = (current_high - current_price) > min_reversal
        bear_reversal = (current_price - current_low) > min_reversal

        if (height_bool and height > 0 and bull_reversal) or (height_bool and height < 0 and bear_reversal):
            # new candle setup
            prev_candle = candle
            candle += 1
            height = 0
            prices_pool = []

        # reversal_control.append((max(prices_pool, default=current_open) - current_price, min(prices_pool, default=current_open) - current_price))

    candles = pd.Series(data=candle_ids, index=trades.index, name='Candle')
    highs = pd.Series(data=high, index=trades.index, name='High')
    lows = pd.Series(data=low, index=trades.index, name='Low')
    opens = pd.Series(data=open_, index=trades.index, name='Open')
    closes = pd.Series(data=close, index=trades.index, name='Close')
    # control = pd.Series(data=reversal_control, index=trades.index, name='control')

    # data = [candles, opens, highs, lows, closes, control]
    # data = [candles*10**-decimal_positions, opens*10**-decimal_positions, highs*10**-decimal_positions, lows*10**-decimal_positions,
    #         closes*10**-decimal_positions, trades['Quantity'], trades['Timestamp']]
    data = [candles, opens, highs, lows, closes, trades['Quantity'], trades['Timestamp']]

    ret = pd.concat(data, axis=1, keys=[s.name for s in data])

    klines = ret.groupby(['Candle']).agg(
        {'Open': 'first', 'High': 'last', 'Low': 'last', 'Close': 'last', 'Quantity': 'sum', 'Timestamp': 'first'})

    date_index = klines['Timestamp'].apply(convert_milliseconds_to_time_zone_datetime, timezoned=time_zone)
    klines.set_index(date_index, inplace=True)

    repair_decimals = np.float(10**-decimal_positions)
    klines['High'] *= repair_decimals
    klines['Low'] *= repair_decimals
    klines['Open'] *= repair_decimals
    klines['Close'] *= repair_decimals

    return klines



import logging
from logging.handlers import RotatingFileHandler

from os import makedirs, path


class Logs:
    def __init__(self, filename='binpan.log', name="BinPan Logger", info_level='DEBUG'):
        try:
            makedirs('logs', exist_ok=True)
        except PermissionError:  # linux
            home = path.expanduser('~')
            makedirs(f'{home}/deploy/deeplab_working', exist_ok=True)
        self.logger = logging.getLogger(name)

        # avoid duplicated logs
        if self.logger.hasHandlers():
            self.logger.handlers = []
            
        self.log_file = filename

        # Create handlers
        self.screen_handler = logging.StreamHandler()
        self.file_handler = logging.FileHandler(self.log_file)
        # self.file_handler = RotatingFileHandler(self.log_file,
        #                                         mode='a',
        #                                         maxBytes=2 * 1024 * 1024,
        #                                         backupCount=5)

        # Create formatters and add it to handlers
        # self.line_format = '%(asctime)s, %(module)10s %(funcName)20s, %(lineno)5d, %(levelname)8s, %(message)s'
        self.line_format = '%(asctime)s %(levelname)8s %(message)s'
        self.screen_format = logging.Formatter(self.line_format, datefmt='%Y-%m-%d\t %H:%M:%S')
        self.file_format = logging.Formatter(self.line_format, datefmt='%Y-%m-%d\t %H:%M:%S')
        self.screen_handler.setFormatter(self.screen_format)
        self.file_handler.setFormatter(self.file_format)

        # Add handlers to the logger
        self.logger.addHandler(self.screen_handler)
        self.logger.addHandler(self.file_handler)

        # Set level
        self.level = eval(f"logging.{info_level}")
        self.logger.setLevel(self.level)

        # return self.logger

    def debug(self, msg):
        self.logger.debug(msg)

    def info(self, msg):
        self.logger.info(msg)

    def warning(self, msg):
        self.logger.warning(msg)

    def error(self, msg):
        self.logger.error(msg, exc_info=True)

    def critical(self, msg):
        self.logger.critical(msg, exc_info=True)



import pandas as pd
from decimal import Decimal as dd
from time import sleep

from .logs import Logs
from .quest import api_raw_signed_get, api_raw_signed_post
from .time_helper import convert_milliseconds_to_str
from .time_helper import convert_string_to_milliseconds
from .market import get_prices_dic, convert_coin

wallet_logger = Logs(filename='./logs/wallet_logger.log', name='wallet_logger', info_level='INFO')


##########
# Helper #
##########

def convert_str_date_to_ms(date: str or int,
                           time_zone: str):
    """
    Converts dates strings formatted as "2022-05-11 06:45:42" to timestamp in milliseconds. If timestamp passed
    it returns the timestamp.

    :param str or int date: Date to check format.
    :param time_zone: A time zone like 'Europe/Madrid'
    :return int: Milliseconds of timestamp.
    """
    if type(date) == str:
        date = convert_string_to_milliseconds(date, timezoned=time_zone)
    return date


##########
# WALLET #
##########


def daily_account_snapshot(account_type: str,
                           decimal_mode: bool,
                           api_key: str,
                           api_secret: str,
                           startTime: int or str = None,
                           endTime: int = None,
                           limit=30,
                           time_zone=None) -> pd.DataFrame:
    """
    The query time period must be inside the previous 30 days.
    Support query within the last month, one month only.
    If startTime and endTime not sent, return records of the last days by default.

    Weight(IP): 2400

    :param str account_type: SPOT or MARGIN (cross)
    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param int limit: Days limit. Default 30.
    :param int or str startTime: Can be integer timestamp in milliseconds or formatted string: 2022-05-11 06:45:42
    :param int or str endTime: Can be integer timestamp in milliseconds or formatted string: 2022-05-11 06:45:42
    :param str time_zone: A time zone to parse index like 'Europe/Madrid'
    :return pd.DataFrame: Pandas DataFrame

    """
    startTime = convert_str_date_to_ms(date=startTime, time_zone=time_zone)
    endTime = convert_str_date_to_ms(date=endTime, time_zone=time_zone)

    available = ['SPOT', 'MARGIN']
    if account_type not in available:
        exc = f"BinPan error: {account_type} not in {available}"
        wallet_logger.error(exc)
        raise Exception(exc)
    elif limit > 30 or limit < 7:
        exc = f"BinPan error: {limit} is over the maximum 30 or under the minimum 7."
        wallet_logger.error(exc)
        raise Exception(exc)

    ret = api_raw_signed_get(endpoint='/sapi/v1/accountSnapshot',
                             params={'type': account_type,
                                     'startTime': startTime,
                                     'endTime': endTime,
                                     'limit': limit},
                             weight=2400, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    rows = []
    for update in ret['snapshotVos']:
        updateTime = update['updateTime']
        data = update['data']
        totalAssetOfBtc = float(data['totalAssetOfBtc'])
        if account_type == 'MARGIN':
            assets_field = 'userAssets'
        else:
            assets_field = 'balances'
        for balance in data[assets_field]:
            free = float(balance['free'])
            locked = float(balance['locked'])
            if free > 0 or locked > 0:
                row = {'updateTime': updateTime,
                       'totalAssetOfBtc': totalAssetOfBtc,
                       'asset': balance['asset'],
                       'free': free,
                       'locked': locked}
                if time_zone:
                    row['datetime'] = convert_milliseconds_to_str(ms=updateTime, timezoned=time_zone)
                rows.append(row)

    if time_zone:
        ret = pd.DataFrame(rows)
        if not ret.empty:
            ret = ret.set_index('datetime').sort_index()
            ret.index.name = f"{account_type} {time_zone}"
    else:
        ret = pd.DataFrame(rows)
        if not ret.empty:
            ret = ret.set_index('updateTime').sort_index()
            ret.index.name = f"{account_type} timestamp"
    return ret


def assets_convertible_dust(decimal_mode: bool,
                            api_key: str,
                            api_secret: str) -> dict:
    """
    Assets dust that can be converted to BNB.
    Weight(IP): 1

    :param bool decimal_mode: It flags to work in decimal mode.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :return dict: A dictionary.

    .. code-block::

       {
            "details": [
                {
                    "asset": "ADA",
                    "assetFullName": "ADA",
                    "amountFree": "6.21",   //Convertible amount
                    "toBTC": "0.00016848",  //BTC amount
                    "toBNB": "0.01777302",  //BNB amount（Not deducted commission fee）
                    "toBNBOffExchange": "0.01741756", //BNB amount（Deducted commission fee）
                    "exchange": "0.00035546" //Commission fee
                }
            ],
            "totalTransferBtc": "0.00016848",
            "totalTransferBNB": "0.01777302",
            "dribbletPercentage": "0.02"     //Commission fee
        }

    """
    yn = input(f"This command may change or convert assets from your wallet!!! are you sure? (y/n)")
    if yn.upper().startswith('Y'):
        return api_raw_signed_post(endpoint='/sapi/v1/asset/dust-btc',
                                   weight=1,
                                   decimal_mode=decimal_mode,
                                   api_key=api_key,
                                   api_secret=api_secret)
    else:
        wallet_logger.warning("Canceled!")


##########
# trades #
##########


def get_fills_price(original_order_dict: dict,
                    isBuyer: bool,
                    margin: bool,
                    test_mode: bool,
                    operation_time: int,
                    decimal_mode: bool,
                    api_key: str,
                    api_secret: str) -> float:
    """
    Obtain averaged price from order API response or claims last trade.

    :param dict original_order_dict: API order dict response.
    :param bool isBuyer: Sets if was a buyer order to filter for wanted order.
    :param bool margin: Sets margin order if it is to claim.
    :param bool test_mode: Set if test mode is on.
    :param int operation_time: Time limit of operation to analyze.
    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :return float: Price averaged.

    Example of not totally done order:

    .. code-block::

        {'symbol': 'ACMBUSD',
        'orderId': 53255492,
        'orderListId': -1,
        'clientOrderId': 'DUHlcCqIX2OO298XoAbvEB',
        'transactTime': 1660332903950,
        'price': 4.158,
        'origQty': 21.1,
        'executedQty': 0.0,
        'cummulativeQuoteQty': 0.0,
        'status': 'NEW',
        'timeInForce': 'GTC',
        'type': 'LIMIT',
        'side': 'BUY',
        'fills': []}


    Example of full SPOT response:

    .. code-block::

        {
          "symbol": "BTCUSDT",
          "orderId": 28,
          "orderListId": -1, //Unless OCO, value will be -1
          "clientOrderId": "6gCrw2kRUFF9CvJDWP16IP",
          "transactTime": 1507725176595,
          "price": "0.00000000",
          "origQty": "10.00000000",
          "executedQty": "10.00000000",
          "cummulativeQuoteQty": "10.00000000",
          "status": "FILLED",
          "timeInForce": "GTC",
          "type": "MARKET",
          "side": "SELL",
          "strategyId": 1,               // This is only visible if the field was populated on order placement.
          "strategyType": 1000000        // This is only visible if the field was populated on order placement.
          "fills": [
            {
              "price": "4000.00000000",
              "qty": "1.00000000",
              "commission": "4.00000000",
              "commissionAsset": "USDT",
              "tradeId": 56
            },
            {
              "price": "3999.00000000",
              "qty": "5.00000000",
              "commission": "19.99500000",
              "commissionAsset": "USDT",
              "tradeId": 57
            },
            {
              "price": "3998.00000000",
              "qty": "2.00000000",
              "commission": "7.99600000",
              "commissionAsset": "USDT",
              "tradeId": 58
            },
            {
              "price": "3997.00000000",
              "qty": "1.00000000",
              "commission": "3.99700000",
              "commissionAsset": "USDT",
              "tradeId": 59
            },
            {
              "price": "3995.00000000",
              "qty": "1.00000000",
              "commission": "3.99500000",
              "commissionAsset": "USDT",
              "tradeId": 60
            }
          ]
        }

    Example of FULL MARGIN RESPONSE:

    .. code-block::

        {
          "symbol": "BTCUSDT",
          "orderId": 28,
          "clientOrderId": "6gCrw2kRUFF9CvJDWP16IP",
          "transactTime": 1507725176595,
          "price": "1.00000000",
          "origQty": "10.00000000",
          "executedQty": "10.00000000",
          "cummulativeQuoteQty": "10.00000000",
          "status": "FILLED",
          "timeInForce": "GTC",
          "type": "MARKET",
          "side": "SELL",
          "marginBuyBorrowAmount": 5,       // will not return if no margin trade happens
          "marginBuyBorrowAsset": "BTC",    // will not return if no margin trade happens
          "isIsolated": true,       // if isolated margin
          "fills": [
            {
              "price": "4000.00000000",
              "qty": "1.00000000",
              "commission": "4.00000000",
              "commissionAsset": "USDT"
            },
            {
              "price": "3999.00000000",
              "qty": "5.00000000",
              "commission": "19.99500000",
              "commissionAsset": "USDT"
            },
            {
              "price": "3998.00000000",
              "qty": "2.00000000",
              "commission": "7.99600000",
              "commissionAsset": "USDT"
            },
            {
              "price": "3997.00000000",
              "qty": "1.00000000",
              "commission": "3.99700000",
              "commissionAsset": "USDT"
            },
            {
              "price": "3995.00000000",
              "qty": "1.00000000",
              "commission": "3.99500000",
              "commissionAsset": "USDT"
            }
          ]
        }

    STOP LIMIT EXAMPLE requested order response:

    .. code-block::

         {'symbol': 'CHZBUSD',
          'orderId': 212320642,
          'orderListId': 72689217,
          'clientOrderId': 'FLc7AHuuZCpChiH99eHtZh',
          'price': 0.2139,
          'origQty': 224.0,
          'executedQty': 224.0,
          'cummulativeQuoteQty': 47.9584,
          'status': 'FILLED',
          'timeInForce': 'GTC',
          'type': 'STOP_LOSS_LIMIT',
          'side': 'SELL',
          'stopPrice': 0.2141,
          'icebergQty': '0.00000000',
          'time': 1661203865507,
          'updateTime': 1661203966780,
          'isWorking': True,
          'origQuoteOrderQty': 0.0},

    """
    ordered = original_order_dict.copy()

    if decimal_mode:
        my_type = dd
    else:
        my_type = float

    if test_mode and 'STOP_LOSS' in ordered['type']:
        return my_type(ordered['stopPrice'])

    if test_mode and 'price' in ordered.keys() and ordered['price']:
        # MUST RETURN FOR TEST ORDERS
        return my_type(ordered['price'])

    if 'fills' in ordered.keys():
        if type(ordered['fills']) == dict:
            if ordered['fills']['price']:
                return my_type(ordered['fills']['price'])

        elif type(ordered['fills']) == list:
            n = 0
            m = 0
            for fill in ordered['fills']:
                price = my_type(fill['price'])
                qty = my_type(fill['qty'])
                n += price * qty
                m += qty
            if m:
                return n / m
        else:
            raise Exception(f"BinPan Error: Error en parseo de orden ejecutada.\n{ordered}")

    # any other case wait 3 times for order to appear
    fills_price = None
    tour = 0
    symbol = ordered['symbol']

    while not fills_price and tour < 5:
        wallet_logger.info(f"Get fills from API trades!... wait 5 seconds for trade to appear. Round:{tour}")

        if not margin:
            last_trades = get_spot_trades_list(symbol=symbol, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
        else:
            last_trades = get_margin_trades_list(symbol=symbol, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)

        try:
            # trades return time field with a timestamp
            fills_price = [i for i in last_trades if i['isBuyer'] == isBuyer and int(i['time']) >= operation_time][-1]['price']
            tour += 1
        except IndexError:
            sleep(5)

    return my_type(fills_price)


def get_fills_qty(original_order_dict: dict,
                  api_key: str,
                  api_secret: str,
                  margin: bool,
                  isBuyer: bool,
                  operation_time: int,
                  decimal_mode: bool) -> float or dd:
    """
    Extracts order quantity from an order response or checks a trade from API.

    :param dict original_order_dict: A putting order response from API dict.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param bool margin: Checks if margin or not.
    :param bool isBuyer: Sets if buy or sell side.
    :param int operation_time: A limit old timestamp to verify existence of order.
    :param bool decimal_mode: Sets Decimal operating mode.
    :return float or decimal.Decimal: Returns the needed value.
    """
    ordered = original_order_dict.copy()
    wallet_logger.debug(f"fills_qty: {ordered}")

    ret = None
    if decimal_mode:
        my_type = dd
    else:
        my_type = float

    if 'fills' in ordered.keys():

        if type(ordered['fills']) == dict:
            return my_type(ordered['fills']['qty'])

        elif type(ordered['fills']) == list:
            m = 0
            for fill in ordered['fills']:
                qty = my_type(fill['qty'])
                m += qty
            return m
        else:
            raise Exception(f"BinPan Error: Error parsing order.\n{ordered}")

    else:
        # si fue una orden de compra a precio de mercado no tiene precio, price = 0.0, cantidad ?????
        if 'test_quantity' in ordered.keys():
            ret = ordered['test_quantity']
        elif 'executedQty' in ordered.keys():
            qty_ret = my_type(ordered['executedQty'])
            if qty_ret != 0:
                return qty_ret
        else:
            fills_qty = None
            tour = 0
            symbol = ordered['symbol']

            while not fills_qty and tour < 5:

                wallet_logger.info(f"Get fills qty from API trades!... wait 5 seconds for trade to appear. Round:{tour}")

                if not margin:
                    last_trades = get_spot_trades_list(symbol=symbol, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
                else:
                    last_trades = get_margin_trades_list(symbol=symbol, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)

                try:
                    # trades return time field with a timestamp
                    fills_qty = [i for i in last_trades if i['isBuyer'] == isBuyer and int(i['time']) >= operation_time][-1]['qty']
                    tour += 1
                except IndexError:
                    sleep(5)
            ret = fills_qty

    return my_type(ret)


def get_spot_trades_list(symbol: str,
                         decimal_mode: bool,
                         api_key: str,
                         api_secret: str,
                         limit: int = 1000,
                         orderId: int = None,
                         startTime: int = None,
                         endTime: int = None,
                         fromId: int = None,
                         recvWindow: int = 10000) -> list:
    """
    Get trades for a specific account and symbol.

    Weight(IP): 10

    :param str symbol: Symbol's trades.
    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param int limit: Default 500; max 1000.
    :param fromId: TradeId to fetch from. Default gets most recent trades. If fromId is set, it will get id >= that fromId. Otherwise,
       most recent trades are returned.
    :param int endTime: Optional.
    :param int startTime: Optional.
    :param int orderId: This can only be used in combination with symbol.
    :param int recvWindow: The value cannot be greater than 60000
    :return list: A list.

    Example:

           .. code-block::

                     [
                        {
                            "symbol": "BNBBTC",
                            "id": 28457,
                            "orderId": 100234,
                            "orderListId": -1, //Unless OCO, the value will always be -1
                            "price": "4.00000100",
                            "qty": "12.00000000",
                            "quoteQty": "48.000012",
                            "commission": "10.10000000",
                            "commissionAsset": "BNB",
                            "time": 1499865549590,
                            "isBuyer": true,
                            "isMaker": false,
                            "isBestMatch": true
                          }
                       ]
    """
    endpoint = '/api/v3/myTrades'
    return api_raw_signed_get(endpoint=endpoint,
                              params={'limit': limit,
                                      'symbol': symbol,
                                      'orderId': orderId,
                                      'startTime': startTime,
                                      'endTime': endTime,
                                      'fromId': fromId,
                                      'recvWindow': recvWindow},
                              weight=10, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)


def get_margin_trades_list(symbol: str,
                           decimal_mode: bool,
                           api_key: str,
                           api_secret: str,
                           isIsolated: bool = False,
                           limit: int = 1000,
                           orderId: int = None,
                           startTime: int = None,
                           endTime: int = None,
                           fromId: int = None,
                           recvWindow: int = 10000) -> list:
    """
    Get margin trades for a specific account and symbol.

    If fromId is set, it will get trades >= that fromId. Otherwise most recent trades are returned.

    Weight(IP): 10

    :param str symbol: Symbol's trades.
    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param bool isIsolated: Sets for getting isolated or not isolated trades. Default is false.
    :param int limit: Default 500; max 1000.
    :param fromId: TradeId to fetch from. Default gets most recent trades. If fromId is set, it will get id >= that fromId. Otherwise,
       most recent trades are returned.
    :param int endTime: Optional.
    :param int startTime: Optional.
    :param int orderId: This can only be used in combination with symbol.
    :param int recvWindow: The value cannot be greater than 60000
    :return list: A list.

    Example:

           .. code-block::

                     [
                        {
                            "commission": "0.00006000",
                            "commissionAsset": "BTC",
                            "id": 34,
                            "isBestMatch": true,
                            "isBuyer": false,
                            "isMaker": false,
                            "orderId": 39324,
                            "price": "0.02000000",
                            "qty": "3.00000000",
                            "symbol": "BNBBTC",
                            "isIsolated": false,
                            "time": 1561973357171
                        }
                    ]
    """

    endpoint = '/sapi/v1/margin/myTrades'

    return api_raw_signed_get(endpoint=endpoint,
                              params={'limit': limit,
                                      'isIsolated': isIsolated,
                                      'symbol': symbol,
                                      'orderId': orderId,
                                      'startTime': startTime,
                                      'endTime': endTime,
                                      'fromId': fromId,
                                      'recvWindow': recvWindow},
                              weight=10, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)


############
# balances #
############


def get_spot_account_info(decimal_mode: bool,
                          api_key: str,
                          api_secret: str,
                          recvWindow: int = 10000) -> dict:
    """
    Get current account information.

    Weight(IP): 10

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param int recvWindow: The value cannot be greater than 60000
    :return dict: A dictionary with data.

    Example:

        .. code-block::

           {
              "makerCommission": 15,
              "takerCommission": 15,
              "buyerCommission": 0,
              "sellerCommission": 0,
              "canTrade": true,
              "canWithdraw": true,
              "canDeposit": true,
              "updateTime": 123456789,
              "accountType": "SPOT",
              "balances": [
                {"asset": "BTC", "free": "4723846.89208129","locked": "0.00000000"},
                {"asset": "LTC","free": "4763368.68006011","locked": "0.00000000"}
              ],
              "permissions": ["SPOT"]
           }

    """
    endpoint = '/api/v3/account'
    return api_raw_signed_get(endpoint=endpoint,
                              params={'recvWindow': recvWindow},
                              weight=10,
                              decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)


def get_spot_free_balances(decimal_mode: bool,
                           api_key: str,
                           api_secret: str,
                           data_dic: dict = None
                           ) -> dict:
    """
    Parses available balances from account info.

    :param dict data_dic: If available, account info can be passed as data_dic parameter to avoid API calling.
    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :return dict: Free balances.
    """
    ret = {}
    if not data_dic:
        data_dic = get_spot_account_info(decimal_mode=decimal_mode, api_secret=api_secret, api_key=api_key)
    wallet_logger.debug(f"spot_free_balances_parsed len(get_spot_account_info()):{len(data_dic)}")
    if decimal_mode:
        for asset in data_dic['balances']:
            qty = dd(asset['free'])
            if qty:
                ret.update({asset['asset']: qty})
    else:
        for asset in data_dic['balances']:
            qty = float(asset['free'])
            if qty:
                ret.update({asset['asset']: qty})
    return ret


def get_spot_locked_balances(decimal_mode: bool,
                             api_key: str,
                             api_secret: str,
                             data_dic: dict = None) -> dict:
    """
    Parses locked in order balances from account info.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param dict data_dic: If available, account info can be passed as data_dic parameter to avoid API calling.
    :return dict: Locked balances.
    """
    ret = {}
    if not data_dic:
        data_dic = get_spot_account_info(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    wallet_logger.debug(f"spot_locked_balances_parsed get_spot_account_info: {len(data_dic)}")

    for asset in data_dic['balances']:
        qty = float(asset['locked'])
        if qty:
            ret.update({asset['asset']: qty})
    return ret


def get_coins_with_balance(decimal_mode: bool, api_key: str, api_secret: str) -> list:
    """
    Get the non-zero balances of an account, free or locked ones. Useful getting wallet value.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :return list:
    """
    data_dic = get_spot_account_info(decimal_mode=decimal_mode, api_secret=api_secret, api_key=api_key)
    free = get_spot_free_balances(data_dic=data_dic, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    locked = get_spot_locked_balances(data_dic=data_dic, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    free = [symbol for symbol, balance in free.items() if float(balance) > 0]
    locked = [symbol for symbol, balance in locked.items() if float(balance) > 0]
    symbols = free + locked
    return list(set(symbols))


def get_spot_balances_df(decimal_mode: bool,
                         api_key: str,
                         api_secret: str,
                         filter_empty: bool = True) -> pd.DataFrame:
    """
    Create a dataframe with the free or blocked amounts of the spot wallet. The index is the assets list.

    Example:

    .. code-block:: python

       from handlers.messages import get_spot_balances_df

       df = get_spot_balances_df()

       print(df)

                   free      locked
            asset
            BNB    0.128359     0.0
            BUSD   0.000001     0.0

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param bool filter_empty: Discards empty quantities.
    :return pd.DataFrame: A dataframe with assets locked or free.
    """
    balances = get_spot_account_info(decimal_mode=decimal_mode,
                                     api_secret=api_secret,
                                     api_key=api_key)['balances']
    df_ = pd.DataFrame(balances)
    df_['free'] = df_['free'].apply(pd.to_numeric)
    df_['locked'] = df_['locked'].apply(pd.to_numeric)
    df_.set_index('asset', drop=True, inplace=True)
    if filter_empty:
        return df_[(df_['free'] != 0) | (df_['locked'] != 0)]
    else:
        return df_


def get_spot_balances_total_value(decimal_mode: bool,
                                  api_key: str,
                                  api_secret: str,
                                  balances_df: pd.DataFrame = None,
                                  convert_to: str = 'BUSD') -> float:
    """
    Returns total value expressed in a quote coin. Counts free and locked assets.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param pd.DataFrame balances_df: A BinPan balances dataframe.
    :param str convert_to: A Binance coin.
    :return float: Total quantity expressed in quote.
    """
    if decimal_mode:
        my_type = dd
    else:
        my_type = float

    if type(balances_df) != pd.DataFrame:
        balances_df = get_spot_balances_df(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)

    prices = get_prices_dic(decimal_mode=decimal_mode)

    symbols = list(balances_df.index)
    free_values = balances_df['free'].tolist()
    locked_values = balances_df['locked'].tolist()
    total = 0

    for i in range(len(symbols)):
        coin = symbols[i]
        free = my_type(free_values[i])
        locked = my_type(locked_values[i])

        if free:
            free = convert_coin(coin=coin,
                                prices=prices,
                                convert_to=convert_to,
                                coin_qty=free,
                                decimal_mode=decimal_mode)

        if locked:
            locked = convert_coin(coin=coin,
                                  prices=prices,
                                  convert_to=convert_to,
                                  coin_qty=locked,
                                  decimal_mode=decimal_mode)
        total += my_type(free)
        total += my_type(locked)

    return total


#################
# MARGIN WALLET #
#################


def get_margin_account_details(decimal_mode: bool, api_key: str, api_secret: str) -> dict:
    """
    Query Cross Margin Account Details (USER_DATA)

    GET /sapi/v1/margin/account (HMAC SHA256)

    Weight(IP): 10

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.

    Response:

    .. code-block::

       {
              "borrowEnabled": true,
              "marginLevel": "11.64405625",
              "totalAssetOfBtc": "6.82728457",
              "totalLiabilityOfBtc": "0.58633215",
              "totalNetAssetOfBtc": "6.24095242",
              "tradeEnabled": true,
              "transferEnabled": true,
              "userAssets": [
                  {
                      "asset": "BTC",
                      "borrowed": "0.00000000",
                      "free": "0.00499500",
                      "interest": "0.00000000",
                      "locked": "0.00000000",
                      "netAsset": "0.00499500"
                  },
                  {
                      "asset": "BNB",
                      "borrowed": "201.66666672",
                      "free": "2346.50000000",
                      "interest": "0.00000000",
                      "locked": "0.00000000",
                      "netAsset": "2144.83333328"
                  },
                  {
                      "asset": "ETH",
                      "borrowed": "0.00000000",
                      "free": "0.00000000",
                      "interest": "0.00000000",
                      "locked": "0.00000000",
                      "netAsset": "0.00000000"
                  },
                  {
                      "asset": "USDT",
                      "borrowed": "0.00000000",
                      "free": "0.00000000",
                      "interest": "0.00000000",
                      "locked": "0.00000000",
                      "netAsset": "0.00000000"
                  }
              ]
        }

    """
    margin_endpoint = '/sapi/v1/margin/account'
    ret = api_raw_signed_get(endpoint=margin_endpoint,
                             params={},
                             weight=10, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    return ret


###################
# margin balances #
###################

def get_margin_balances(decimal_mode: bool, api_key: str, api_secret: str) -> dict:
    """
    Collects balances in the margin account that are not null.

    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :return dict: A dictionary with coins with balances.

    Example:

    .. code-block::

       {'BNB': {'free': 0.06,
          'locked': 0.0,
          'borrowed': 0.0,
          'interest': 0.0,
          'netAsset': 0.06},
         'BUSD': {'free': 50.0,
          'locked': 0.0,
          'borrowed': 0.0,
          'interest': 0.0,
          'netAsset': 50.0}}

    """
    margin_status = get_margin_account_details(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    ret = {}
    for asset in margin_status['userAssets']:
        entry = {}
        for k, v in asset.items():
            if k != 'asset':
                if decimal_mode:
                    entry[k] = dd(v)
                else:
                    entry[k] = float(v)
        if sum(entry.values()) != 0:
            ret[asset['asset']] = entry
    return ret


def get_margin_free_balances(decimal_mode: bool,
                             api_key: str,
                             api_secret: str,
                             balances: dict = None
                             ) -> dict:
    """
    Just returns free existing balances. It is optional to avoid an API call.

    :param dict balances: Returns dict with assets as keys and a float value for not null quantities.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param bool decimal_mode: Fixes Decimal return type.
    :return dict: A dict with float values.

    """
    if not balances:
        balances = get_margin_balances(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    ret = {}
    for k, v in balances.items():
        if 'free' in v.keys():
            if v['free']:
                ret[k] = v['free']
    return ret


def get_margin_locked_balances(decimal_mode: bool,
                               api_key: str,
                               api_secret: str,
                               balances: dict = None) -> dict:
    """
    Just returns locked existing balances. It is optional to avoid an API call.

    :param bool decimal_mode: Fixes Decimal return type.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param dict balances: Returns dict with assets as keys and a float value for not null quantities.
    :return dict: A dict with float values.

    """
    if not balances:
        balances = get_margin_balances(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    ret = {}
    for k, v in balances.items():
        if 'locked' in v.keys():
            if v['locked']:
                ret[k] = v['locked']
    return ret


def get_margin_borrowed_balances(decimal_mode: bool,
                                 api_key: str,
                                 api_secret: str,
                                 balances: dict = None) -> dict:
    """
    Just returns borrowed existing balances. It is optional to avoid an API call.

    :param bool decimal_mode: Fixes Decimal return type.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param dict balances: Returns dict with assets as keys and a float value for not null quantities.
    :return dict: A dict with float values.

    """
    if not balances:
        balances = get_margin_balances(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    ret = {}
    for k, v in balances.items():
        if 'borrowed' in v.keys():
            if v['borrowed']:
                ret[k] = v['borrowed']
    return ret


def get_margin_interest_balances(decimal_mode: bool,
                                 api_key: str,
                                 api_secret: str,
                                 balances: dict = None) -> dict:
    """
    Just returns interest existing balances. It is optional to avoid an API call.

    :param bool decimal_mode: Fixes Decimal return type.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param dict balances: Returns dict with assets as keys and a float value for not null quantities.
    :return dict: A dict with float values.

    """
    if not balances:
        balances = get_margin_balances(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    if not balances:
        balances = get_margin_balances(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    ret = {}
    for k, v in balances.items():
        if 'interest' in v.keys():
            if v['interest']:
                ret[k] = v['interest']
    return ret


def get_margin_netAsset_balances(decimal_mode: bool,
                                 api_key: str,
                                 api_secret: str,
                                 balances: dict = None):
    """
    Just returns netAsset existing balances. It is optional to avoid an API call.

    :param bool decimal_mode: Fixes Decimal return type.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param dict balances: Returns dict with assets as keys and a float value for not null quantities.
    :return dict: A dict with float values.

    """
    if not balances:
        balances = get_margin_balances(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    return {k: v['netAsset'] for k, v in balances.items() if v['netAsset']}


def get_margin_balances_total_value(decimal_mode: bool,
                                    api_key: str,
                                    api_secret: str,
                                    balances: dict = None,
                                    convert_to: str = 'BUSD'
                                    ) -> float or dd:
    """
    Returns total value expressed in a quote coin. Counts free, locked, borrowed and interest assets.

    :param dict balances: A BinPan balances dict. It is optional to avoid an API call.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param str convert_to: A Binance coin.
    :param bool decimal_mode: Fixes Decimal return type.
    :return float: Total quantity expressed in quote.
    """
    if decimal_mode:
        my_type = dd
    else:
        my_type = float
    prices = get_prices_dic(decimal_mode=decimal_mode)

    if not balances:
        balances = get_margin_balances(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)

    assets = list(balances.keys())

    free_values = get_margin_free_balances(balances=balances, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    locked_values = get_margin_locked_balances(balances=balances, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    borrowed_values = get_margin_borrowed_balances(balances=balances, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)

    # TODO: saber si el interest tiene signo negativo
    interest_values = get_margin_interest_balances(balances=balances, decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)

    total = 0

    for i in range(len(assets)):
        coin = assets[i]

        if coin in free_values.keys():
            free = free_values[coin]
        else:
            free = 0
        if coin in locked_values.keys():
            locked = locked_values[coin]
        else:
            locked = 0
        if coin in borrowed_values.keys():
            borrowed = borrowed_values[coin]
        else:
            borrowed = 0
        if coin in interest_values.keys():
            interest = interest_values[coin]
        else:
            interest = 0

        # TODO: saber si el interest tiene signo negativo
        total_coins = free + locked + borrowed - interest

        converted_value = convert_coin(coin=coin,
                                       prices=prices,
                                       convert_to=convert_to,
                                       coin_qty=total_coins,
                                       decimal_mode=decimal_mode)

        total += my_type(converted_value)
    return my_type(total)



from .starters import AesCipher
from .logs import Logs
from .time_helper import convert_milliseconds_to_str
from .wallet import get_spot_balances_df, get_spot_balances_total_value

from pandas import DataFrame
import requests

msg_logger = Logs(filename='./logs/msg_logger.log', name='msg_logger', info_level='INFO')

cipher_object = AesCipher()

try:
    from secret import encoded_chat_id, encoded_telegram_bot_id

except Exception as exc:
    msg = """\n\n-------------------------------------------------------------
WARNING: No API Key or API Secret
    
Not found telegram bot key or chat key for the telegram message module.

Example adding bot key and chat id:

    from binpan import handlers

    # from @BotFather, get the bot api key
    handlers.files.add_any_key(key="xxxxxxxxxx", key_name="encoded_telegram_bot_id")

    # write to your bot and then get your chat id from https://api.telegram.org/bot<YourBOTToken>/getUpdates
    handlers.files.add_any_key(key="xxxxxxxxxx", key_name="encoded_chat_id")
    
    
"""
    msg_logger.warning(msg)
    encoded_telegram_bot_id = ''
    encoded_chat_id = ''


def telegram_bot_send_text(msg: dict or str,
                           parse_mode='Markdown&text',
                           disable_notification=False,
                           alt_enc_bot_id: str = None,
                           alt_chat_id: str = None) -> dict:
    """
    Sends a telegram message.

    It takes bot key and chat id from secret encrypted file. It is required to create previously the bot and chat keys with the
    variable names *encoded_chat_id*, *encoded_telegram_bot_id*.

    Example adding bot key and chat id:

    .. code-block:: python

        from binpan import handlers

        # from @BotFather, get the bot api key
        handlers.files.add_any_key(key="xxxxxxxxxx", key_name="encoded_telegram_bot_id")

        # write to your bot and then get your chat id from https://api.telegram.org/bot<YourBOTToken>/getUpdates
        handlers.files.add_any_key(key="xxxxxxxxxx", key_name="encoded_chat_id")


    :param str msg: A message
    :param str parse_mode: Parsing telegram message. Default is 'Markdown&text'.
    :param disable_notification: Avoids getting notified by the telegram message. Only working in group chats.
        https://stackoverflow.com/questions/64360221/disable-notification-in-python-telegram-bot
    :param str alt_enc_bot_id: Optional alternative bot api key.
    :param str alt_chat_id: Optional alternative chat id.
    :return: Telegrams API response.
    """

    if type(msg) == dict:
        bot_message = str(msg.copy())
    else:
        bot_message = str(msg)

    if not alt_enc_bot_id:
        alt_enc_bot_id = cipher_object.decrypt(encoded_telegram_bot_id)
    if not alt_chat_id:
        alt_chat_id = cipher_object.decrypt(encoded_chat_id)

    send_text = f'https://api.telegram.org/bot{alt_enc_bot_id}/sendMessage?chat_id={alt_chat_id}&' \
                f'parse_mode={parse_mode}={bot_message}'

    if disable_notification:
        send_text += "&disable_notification=true"

    response = requests.get(send_text)

    msg_logger.info("TELEGRAM " + " ".join(bot_message.replace('\n', ' ').split()))

    return response.json()


def telegram_parse_dict(msg_data: dict, timezone='UTC'):
    """
    Parses a dict and downcast types.

    :param dict msg_data: Dict to parse as message.
    :param str timezone: A time zone to parse more human readable any field containing "time" in the name.
    :return str: A markdown v1 parsed telegram string.
    """
    msg_dict = msg_data.copy()
    parsed_msg = ""
    for k, v in msg_dict.items():
        try:
            fv1 = float(v)
        except:
            fv1 = v
        try:
            fv2 = int(fv1)
            assert fv2 == fv1
        except:
            fv2 = fv1
        if 'pct' in k:
            fv2 = fv2 * 100
            row = f"*{k}* : `{fv2:.2f}` \n"
        elif type(fv2) == float:
            row = f"*{k}* : `{fv2:.8f}` \n"
        elif type(fv2) == int:
            row = f"*{k}* : `{fv2}` \n"
        elif type(fv2) == dict:
            row = f"*{k}* : \n{telegram_parse_dict(msg_data=fv2, timezone=timezone)} \n"
        else:
            if 'time' in k and k != 'timeInForce' and k != 'time_zone':
                date = convert_milliseconds_to_str(ms=fv2, timezoned=timezone)
                row = f"*{k}* : {date} \n"
            else:
                row = f"*{k}* : {fv2} \n"
        parsed_msg += row
    return parsed_msg.replace('_', ' ').replace("Decimal('", "`").replace("')", "`")


# future bot #


def tab_str(text: str, indentation=8) -> str:
    """
    Indentation for telegram messages.
    :param str text: A message with spaces to insert tabulations in.
    :param int indentation: The spaces quantity to insert.
    :return str: Tabulated message.
    """
    sp = text.split()
    ret = ""
    for i in sp:
        length = len(i)
        ind = (indentation - length) % indentation
        ret += i + " " * ind
    return ret


def telegram_parse_dataframe_markdown(data: DataFrame,
                                      indentation: int = 6,
                                      title: str = "Balances") -> str:
    """
    Parses a Dataframe in telegrams message format.

    :param DataFrame data: A dataframe
    :param int indentation: Spaces to insert each tabulation.
    :param str title: a header or title to the result.
    :return str: String parsed with tabulations.
    """
    ret = f"*{title}*: \n"
    cols = data.columns
    ret += "```\n"
    for idx, row in data.iterrows():
        data_row = row.tolist()
        str_row = f" {idx} " + ''.join([f"{cols[i]}: {d:.8f} " for i, d in enumerate(data_row)])
        ret += str(tab_str(str_row, indentation=indentation)) + "\n"
    return ret + "```"


def telegram_parse_order_markdown(original_order: dict,
                                  isBuyer: bool,
                                  margin: bool,
                                  test_mode: bool,
                                  timezoned='Europe/Madrid'
                                  ):
    """
    Parses API order response to telegram string.

    :param dict original_order: API order response.
    :param bool isBuyer: Sets if was a buyer order to filter for wanted order.
    :param bool margin: Sets margin order if it is to claim.
    :param bool test_mode: Set if test mode is on.
    :param str timezoned: a time zone like 'Europe/Madrid' for human readable timestamps.
    :return str: Parsed telegram style  string.
    """
    order_ = original_order.copy()
    parsed = ""

    if 'orderReports' in order_.keys():
        p_stop = telegram_parse_order_markdown(order_['orderReports'][0], isBuyer=isBuyer, margin=margin, test_mode=test_mode)
        p_limit = telegram_parse_order_markdown(order_['orderReports'][1], isBuyer=isBuyer, margin=margin, test_mode=test_mode)
        stop_parsed = f"\n{p_stop}\n"
        limit_parsed = f"\n{p_limit}\n"
    else:
        stop_parsed = 'Possible error with stop_parsed and orderReports'
        limit_parsed = 'Possible error with and limit_parsed orderReports'

    if 'time' in order_.keys():
        date = convert_milliseconds_to_str(int(order_['time']), timezoned=timezoned)
        order_.update({'time': date})

    if 'transactTime' in order_.keys():
        date = convert_milliseconds_to_str(int(order_['transactTime']), timezoned=timezoned)
        order_.update({'transactTime': date})

    if 'updateTime' in order_.keys():
        date = convert_milliseconds_to_str(int(order_['updateTime']), timezoned=timezoned)
        order_.update({'updateTime': date})

    if 'quantity' in order_.keys():
        order_.update({'quantity': f"{order_['quantity']}"})

    if 'price' in order_.keys():
        order_.update({'price': f"{order_['price']}"})

    if 'origQty' in order_.keys():
        order_.update({'origQty': f"{order_['origQty']}"})

    if 'executedQty' in order_.keys():
        order_.update({'executedQty': f"{order_['executedQty']}"})

    if 'stopPrice' in order_.keys():
        order_.update({'stopPrice': f"{order_['stopPrice']}"})

    if 'fills' in order_.keys():
        order_.update({'fills': f"{order_['fills']}"})

    for key, val in order_.items():
        if key == 'orderReports':
            line = stop_parsed
            parsed += line
            line = limit_parsed
            parsed += line
        elif val:
            line = rf"*{key}* : `{val}` " + "\n"
            parsed += line

    return parsed


def send_balances(decimal_mode: bool,
                  api_key: str,
                  api_secret: str,
                  convert_to: str = 'BUSD') -> float:
    """
    Sends telegram message with total value of spot wallet in selected coin.

    It returns free and locked assets value added.
    :param bool decimal_mode: Fixes Decimal return type and operative.
    :param str api_key: A BinPan encrypted in BinPan's way api key. Do not use unencrypted api values, import it from  secret.py file.
    :param str api_secret: A BinPan encrypted in BinPan's way api secret. Do not use unencrypted api values, import it from  secret.py file.
    :param str convert_to: A Binance coin.
    :return float: total value of wallet.

    """
    balances_df = get_spot_balances_df(decimal_mode=decimal_mode, api_key=api_key, api_secret=api_secret)
    parsed_balances = telegram_parse_dataframe_markdown(balances_df)

    # add total value in usdt
    total_value = get_spot_balances_total_value(balances_df=balances_df,
                                                convert_to=convert_to,
                                                decimal_mode=decimal_mode,
                                                api_key=api_key,
                                                api_secret=api_secret)

    parsed_balances += f"\n*Total value*: `{total_value}` \n*Quote*: `{convert_to}`"
    telegram_bot_send_text(parsed_balances)
    return total_value



from . import exceptions
from . import exchange
from . import files
from . import logs
from . import market
from . import messages
from . import plotting
from . import quest
from . import redis_fetch
from . import starters
from . import strategies
from . import time_helper
from . import wallet
from . import indicators
from . import tags




import json
from .logs import Logs
from .messages import telegram_bot_send_text

exceptions_logger = Logs(filename='./logs/exceptions.log', name='exceptions', info_level='INFO')


############################
# Exceptions
############################

# TODO: color and row exceptions control


class MissingApiData(Exception):
    def __init__(self, message):
        self.message = message
        self.msg = f"""No API Key or API Secret. 
        
        API key would be needed for personal API calls. Any other calls will work.

        Adding:

        from binpan import handlers

        handlers.files.add_api_key("xxxx")
        handlers.files.add_api_secret("xxxx")

        API keys will be added to a file called secret.py in an encrypted way. API keys in memory stay encrypted except in the API call instant.

        Create API keys: https://www.binance.com/en/support/faq/360002502072
        
        Exception: {self.message}
        """

    def __str__(self):
        return self.msg


class BinanceAPIException(Exception):
    """
    Extract exception from response.

    """

    def __init__(self, response, status_code, text):
        self.code = 0
        try:
            json_res = json.loads(text)
        except ValueError:
            self.message = 'Invalid JSON error message from Binance: {}'.format(response.text)
        else:
            self.code = json_res['code']
            self.message = json_res['msg']
        self.status_code = status_code
        self.response = response
        self.request = getattr(response, 'request', None)

    def __str__(self):  # pragma: no cover
        return 'APIError(code=%s): %s' % (self.code, self.message)


class BinanceRequestException(Exception):
    """
    Request methods exceptions.

    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f'BinanceRequestException: {self.message}'


class BinanceOrderException(Exception):
    """
    Orders exceptions from API.

    """

    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __str__(self):
        return 'BinanceOrderException(code=%s): %s' % (self.code, self.message)


class BinanceOrderMinAmountException(BinanceOrderException):
    """
    Symbol filter exceptions: Minimum Amount

    """

    def __init__(self, value):
        message = "Amount must be a multiple of %s" % value
        super().__init__(-1013, message)


class BinanceOrderMinPriceException(BinanceOrderException):
    """
    Symbol filter exceptions: Minimum Price

    """

    def __init__(self, value):
        message = "Price must be at least %s" % value
        super().__init__(-1013, message)


class BinanceOrderMinTotalException(BinanceOrderException):
    """
    Symbol filter exceptions: Minimum Total

    """

    def __init__(self, value):
        message = "Total must be at least %s" % value
        super().__init__(-1013, message)


class BinanceOrderUnknownSymbolException(BinanceOrderException):
    """
    Unknown symbol.

    """

    def __init__(self, value):
        message = "Unknown symbol %s" % value
        super().__init__(-1013, message)


class BinanceOrderInactiveSymbolException(BinanceOrderException):
    """
    Order interactive exception.

    """

    def __init__(self, value):
        message = "Attempting to trade an inactive symbol %s" % value
        super().__init__(-1013, message)


class NotImplementedException(Exception):
    """
    Not implemented.

    """

    def __init__(self, value):
        message = f'Not implemented: {value}'
        super().__init__(message)


#
# class BinPanException(Exception):
#     """
#     Message the exception and logs it.
#     """
#
#     def __init__(self, exception_class, doc, message):
#         self.exception_class = exception_class
#         self.doc = doc
#         self.message = message
#
#     def __str__(self):
#         msg = f'BinPan Exception: {self.exception_class} {self.doc} {self.message}'
#         exceptions_logger.error(msg)
#

class BinPanException(Exception):
    """
    BinPan exception with custom message.

    :param str msg: A message for the Exception message.
    :param bool telegram_send:
    """

    def __init__(self,
                 msg: str,
                 telegram_send: bool = False):
        self.message = msg
        self.telegram_send = telegram_send

        self.internal_msg = f"BinPan Cache Exception: {msg}"

        exceptions_logger.error(msg)

        if telegram_send:
            self.telegram_msg()

        super().__init__(self.message)

    def __str__(self):
        return self.message

    def telegram_msg(self):
        """
        Sends to telegram the internal message.
        """
        telegram_bot_send_text(msg=self.internal_msg)



from typing import List

import handlers.market
from .time_helper import convert_utc_ms_column_to_time_zone, convert_datetime_to_string, convert_milliseconds_to_utc_string, \
    convert_milliseconds_to_time_zone_datetime, convert_milliseconds_to_str, open_from_milliseconds, next_open_by_milliseconds
from .market import tick_seconds
import pandas as pd
import json
from .logs import Logs
from redis import StrictRedis
from time import sleep, time
import numpy as np

redis_logger = Logs(filename='./logs/redis_fetch.log', name='redis_fetch', info_level='INFO')

klines_columns = {"t": "Open timestamp",
                  "o": "Open",
                  "h": "High",
                  "l": "Low",
                  "c": "Close",
                  "v": "Volume",
                  "T": "Close timestamp",
                  "q": "Quote volume",
                  "n": "Trades",
                  "V": "Taker buy base volume",
                  "Q": "Taker buy quote volume",
                  "B": "Ignore"}


##########
# Parser #
##########

def redis_klines_parser(json_list: List[str],
                        symbol: str,
                        tick_interval: str,
                        time_zone: str = 'Europe/madrid',
                        time_index: bool = True,
                        ) -> pd.DataFrame:
    """
    Parses redis klines list to a BinPan Dataframe.

    :param List[str] json_list: A list with klines format data.
    :param str symbol: Symbol expected. Just for naming index.
    :param str tick_interval: Tick interval expected. Just for naming index.
    :param str time_zone:  A time zone for converting the index. Example: 'Europe/Madrid'
    :param bool time_index: If true, index are datetime format, else integer index.
    :return pd.DataFrame: A BinPan dataframe.

    """

    time_cols = ['Open time', 'Close time']
    dicts_data = [json.loads(i) for i in json_list]
    df = pd.DataFrame(data=dicts_data)
    df.rename(columns=klines_columns, inplace=True)

    df = handlers.market.convert_to_numeric(data=df)
    # for col in df.columns:
    #     df[col] = pd.to_numeric(arg=df[col], downcast='integer')

    df.loc[:, 'Open time'] = df['Open timestamp']
    df.loc[:, 'Close time'] = df['Close timestamp']

    if time_zone != 'UTC':  # converts to time zone the time columns
        for col in time_cols:
            df.loc[:, col] = convert_utc_ms_column_to_time_zone(df, col, time_zone=time_zone)
            df.loc[:, col] = df[col].apply(lambda x: convert_datetime_to_string(x))
    else:
        for col in time_cols:
            df.loc[:, col] = df[col].apply(lambda x: convert_milliseconds_to_utc_string(x))

    if time_index:
        date_index = df['Open timestamp'].apply(convert_milliseconds_to_time_zone_datetime, timezoned=time_zone)
        df.set_index(date_index, inplace=True)

    index_name = f"{symbol.upper()} {tick_interval} {time_zone}"
    df.index.name = index_name
    return df[['Open time', 'Open', 'High', 'Low', 'Close', 'Volume', 'Close time', 'Quote volume', 'Trades', 'Taker buy base volume',
               'Taker buy quote volume', 'Ignore', 'Open timestamp', 'Close timestamp']]


###############
# Redis Utils #
###############

def fetch_keys(redisClient: StrictRedis,
               filter_tick_interval=None,
               filter_quote: str = None) -> list:
    """
    Fetch all keys in redis database.

    :param StrictRedis redisClient: A redis connector.
    :param str filter_tick_interval: Optional. A binance klines tick interval to fetch all keys for that interval.
    :param str filter_quote: Filter symbols without a quote.
    :return list: Returns all keys for a tick interval if passed, else all existing keys in redis.
    """
    ret = redisClient.scan_iter()
    if filter_tick_interval:
        ret = [i for i in ret if i.endswith(filter_tick_interval)]
    if filter_quote:
        ret = [i for i in ret if i.split('@')[0].endswith(filter_quote.lower())]
    return list(ret)


def klines_continuity(klines: list, tick_interval: str):
    """
    Check klines missing.

    :param list klines: List of raw candles response in a list of strings, not json parsed yet.
    :param str tick_interval: Binance klines interval.
    :return bool: If true, klines are continuous.
    """
    try:
        start = int(klines[0][6:19])
        end = int(klines[-1][6:19])
        ticks = int((end - start) / (tick_seconds[tick_interval] * 1000))
        assert ticks + 1 == len(klines)
        return True
    except:
        print("Klines not continuously spaced!")
        return False


def klines_ohlc_to_numpy(klines: list) -> tuple:
    """
    Extract open, high, low and close numpy arrays from raw klines (list of json strings).

    :param list klines: A list with json strings.
    :return tuple: A tuple with numpy arrays with data: open, high, low and close.
    """
    open_ = []
    high = []
    low = []
    close = []
    for row in klines:
        json_row = json.loads(row)

        open_.append(json_row['o'])
        high.append(json_row['h'])
        low.append(json_row['l'])
        close.append(json_row['c'])

    ret1 = np.fromiter(open_, dtype=np.float64)
    ret2 = np.fromiter(high, dtype=np.float64)
    ret3 = np.fromiter(low, dtype=np.float64)
    ret4 = np.fromiter(close, dtype=np.float64)

    return ret1, ret2, ret3, ret4


def redis_baliza(redis_client: StrictRedis,
                 symbol: str,
                 tick_interval: str,
                 time_zone='Europe/madrid') -> int:
    """
    Returns when Redis data available waiting until the next open timestamp appears.

    Expected format for the redis data is:

    .. code-block::

       '{"t": 1657651320000, "o": "0.10100000", "h": "0.10100000", "l": "0.10100000", "c": "0.10100000", "v": "0.00000000",
         "T": 1657651379999, "q": "0.00000000", "n": 0, "V": "0.00000000", "Q": "0.00000000", "B": "0"}'


    :param StrictRedis redis_client: A redis client.
    :param str symbol: Symbol.
    :param str tick_interval: Binance tick interval.
    :param str time_zone: A time zone for messages.
    :return int: Timestamp of the open waited for.
    """

    now = int(time() * 1000)
    my_open = open_from_milliseconds(now, tick_interval=tick_interval)
    next_open = next_open_by_milliseconds(now, tick_interval=tick_interval)

    redis_logger.debug(
        f"Baliza {symbol} {tick_interval} started at {convert_milliseconds_to_str(ms=my_open, timezoned=time_zone)} and waiting "
        f"until {convert_milliseconds_to_str(ms=next_open, timezoned=time_zone)} is open.")

    while True:

        now = int(time() * 1000)
        last_redis_ts = fetch_zset_range(redisClient=redis_client,
                                         symbol=symbol.lower(),
                                         tick_interval=tick_interval,
                                         start_index=-1)
        fetched_open_ts = int(last_redis_ts[-1][6:19])

        # redis_logger.debug(f"Fetched {convert_milliseconds_to_str(ms=fetched_open_ts, timezoned=time_zone)}")

        if fetched_open_ts >= my_open:  # because redis candles apper just when closed

            redis_logger.debug(
                f"Fetched at {convert_milliseconds_to_str(ms=now, timezoned=time_zone)} with a late of {(now - next_open) / 1000} seconds.")
            break

        else:
            sleep(0.01)

    return fetched_open_ts


###############
# Redis Zsets #
###############

def push_to_ordered_set(redisClient: StrictRedis,
                        key: str,
                        mapping: dict,
                        LT=False,
                        XX=False,
                        NX=True,
                        GT=False,
                        CH=True,
                        INCR=False,
                        ):
    """
    Pushes elements to the ordered set with a score as index.

    :param StrictRedis redisClient: A redis client connector.
    :param str key: The redis key name.
    :param list mapping: Data to push in the format {data: score}
    :param LT: Only update existing elements if the new score is less than the current score. This flag doesn't prevent adding new elements.
    :param XX: Only update elements that already exist. Don't add new elements.
    :param NX: Only add new elements. Don't update already existing elements. This is the default only true value.
    :param GT: Only update existing elements if the new score is greater than the current score. This flag doesn't prevent adding new
       elements.
    :param CH: Modify the return value from the number of new elements added, to the total number of elements changed (CH is an abbreviation
       of changed). Changed elements are new elements added and elements already existing for which the score was updated. So elements
       specified in the command line having the same score as they had in the past are not counted. Note: normally the return value of ZADD
       only counts the number of new elements added.
    :param INCR: When this option is specified ZADD acts like ZINCRBY. Only one score-element pair can be specified in this mode.
       Note: The GT, LT and NX options are mutually exclusive.
    :return: redis feedback info.

    """
    return redisClient.zadd(name=key, mapping=mapping, lt=LT, xx=XX, nx=NX, gt=GT, ch=CH, incr=INCR)


def fetch_zset_range(redisClient: StrictRedis,
                     key: str = None,
                     symbol=None,
                     tick_interval=None,
                     start_index=0,
                     end_index=-1,
                     with_scores=False) -> list:
    """
    Fetch a redis ordered set by its key name and start end indexes. Optionally returns with redis index (scores) in a tuple each row
    if with_Scores.

    :param StrictRedis redisClient: A redis connector.
    :param str key: Redis key name. Optional, symbol and tick interval can be used too.
    :param str symbol: A Binance valid Symbol.
    :param str tick_interval: A binance tick interval. Like 5m, 1h etc
    :param int start_index: Numeric index in redis key.
    :param int end_index: Numeric index in redis key.
    :param bool with_scores: Will return rows with its index in a tuple if true.
    :return list: A list of rows in the redis set or a list of tuples with rows and scores if with_scores true.


    Example:

    .. code-block::

        from binpan import binpan
        from handlers import redis_fetch
        from random import choice

        redis_client = binpan.redis_client(ip='192.168.69.43')

        keys = redis_fetch.fetch_keys(redisClient=redis_client)
        stream = choice(keys)

        print(stream)

        >>> 'galabusd@kline_5m'

        redis_fetch.fetch_zset_range(redisClient=redis_client, key=stream, start_index=0, end_index=3)

        >>> ['{"t": 1658765700000, "o": "0.05011000", "h": "0.05017000", "l": "0.05004000", "c": "0.05006000", "v": "345741.00000000", "T": 1658765999999, "q": "17315.68030000", "n": 87, "V": "84642.00000000", "Q": "4239.51131000", "B": "0"}',
             '{"t": 1658766000000, "o": "0.05005000", "h": "0.05010000", "l": "0.04999000", "c": "0.04999000", "v": "448270.00000000", "T": 1658766299999, "q": "22422.46695000", "n": 68, "V": "132503.00000000", "Q": "6628.86299000", "B": "0"}',
             '{"t": 1658766300000, "o": "0.04998000", "h": "0.05004000", "l": "0.04995000", "c": "0.05000000", "v": "268084.00000000", "T": 1658766599999, "q": "13396.55626000", "n": 57, "V": "94341.00000000", "Q": "4715.14561000", "B": "0"}',
             '{"t": 1658766600000, "o": "0.04998000", "h": "0.05010000", "l": "0.04996000", "c": "0.05002000", "v": "402674.00000000", "T": 1658766899999, "q": "20162.97689000", "n": 68, "V": "154895.00000000", "Q": "7753.09416000", "B": "0"}']

    """
    if not key:
        key = f"{symbol}@kline_{tick_interval}"
    return redisClient.zrange(name=key, start=start_index, end=end_index, withscores=with_scores)


def fetch_zset_timestamps(redisClient: StrictRedis,
                          key: str,
                          start_timestamp: int,
                          end_timestamp: int = None,
                          with_scores=False) -> list:
    """
    Fetch a redis ordered set by its key name and start end scores. Optionally with redis index (scores) in a tuple each row.
    Optionally returns with redis index (scores) in a tuple each row if with_Scores.

    :param StrictRedis redisClient: A redis connector.
    :param str key: Redis key name.
    :param int start_timestamp: Usually a timestamp as index in redis key.
    :param int end_timestamp: Usually a timestamp as index in redis key.
    :param bool with_scores: Will return rows with its index in a tuple if true.
    :return list: A list of rows in the redis set or a list of tuples with rows and scores if with_scores true.

    Example:

    .. code-block::

        from binpan import binpan
        from handlers import redis_fetch
        from random import choice

        redis_client = binpan.redis_client(ip='192.168.69.43')

        keys = redis_fetch.fetch_keys(redisClient=redis_client)
        stream = choice(keys)

        print(stream)

        >>> 'galabusd@kline_5m'

        redis_fetch.fetch_zset_timestamps(redisClient=redis_client, key=stream, start_timestamp=1658765700000, end_timestamp= 1658766600000, with_scores=True)

        >>> [('{"t": 1658765700000, "o": "0.05011000", "h": "0.05017000", "l": "0.05004000", "c": "0.05006000", "v": "345741.00000000", "T": 1658765999999, "q": "17315.68030000", "n": 87, "V": "84642.00000000", "Q": "4239.51131000", "B": "0"}',
              1658765700000.0),
             ('{"t": 1658766000000, "o": "0.05005000", "h": "0.05010000", "l": "0.04999000", "c": "0.04999000", "v": "448270.00000000", "T": 1658766299999, "q": "22422.46695000", "n": 68, "V": "132503.00000000", "Q": "6628.86299000", "B": "0"}',
              1658766000000.0),
             ('{"t": 1658766300000, "o": "0.04998000", "h": "0.05004000", "l": "0.04995000", "c": "0.05000000", "v": "268084.00000000", "T": 1658766599999, "q": "13396.55626000", "n": 57, "V": "94341.00000000", "Q": "4715.14561000", "B": "0"}',
              1658766300000.0),
             ('{"t": 1658766600000, "o": "0.04998000", "h": "0.05010000", "l": "0.04996000", "c": "0.05002000", "v": "402674.00000000", "T": 1658766899999, "q": "20162.97689000", "n": 68, "V": "154895.00000000", "Q": "7753.09416000", "B": "0"}',
              1658766600000.0)]

        # Now without scores

        redis_fetch.fetch_zset_timestamps(redisClient=redis_client, key=stream, start_timestamp=1658765700000, end_timestamp= 1658766600000, with_scores=False)

        >>> ['{"t": 1658765700000, "o": "0.05011000", "h": "0.05017000", "l": "0.05004000", "c": "0.05006000", "v": "345741.00000000", "T": 1658765999999, "q": "17315.68030000", "n": 87, "V": "84642.00000000", "Q": "4239.51131000", "B": "0"}',
            '{"t": 1658766000000, "o": "0.05005000", "h": "0.05010000", "l": "0.04999000", "c": "0.04999000", "v": "448270.00000000", "T": 1658766299999, "q": "22422.46695000", "n": 68, "V": "132503.00000000", "Q": "6628.86299000", "B": "0"}',
            '{"t": 1658766300000, "o": "0.04998000", "h": "0.05004000", "l": "0.04995000", "c": "0.05000000", "v": "268084.00000000", "T": 1658766599999, "q": "13396.55626000", "n": 57, "V": "94341.00000000", "Q": "4715.14561000", "B": "0"}',
            '{"t": 1658766600000, "o": "0.04998000", "h": "0.05010000", "l": "0.04996000", "c": "0.05002000", "v": "402674.00000000", "T": 1658766899999, "q": "20162.97689000", "n": 68, "V": "154895.00000000", "Q": "7753.09416000", "B": "0"}']

    """
    ret = redisClient.zrangebyscore(name=key,
                                    min=start_timestamp,
                                    max=end_timestamp,
                                    withscores=with_scores)
    return ret


def fetch_set_and_parse(redisClient: StrictRedis,
                        key: str) -> pd.DataFrame:
    """
    Fetch a websocket channel from redis and parses to a BinPan dataframe.

    Data is expected to be in a redis ordered set with klines data each row and with Open timestamp as score.
    Example of data in one score of the zset:

    .. code-block::

       '{"t": 1657651320000, "o": "0.10100000", "h": "0.10100000", "l": "0.10100000", "c": "0.10100000", "v": "0.00000000",
         "T": 1657651379999, "q": "0.00000000", "n": 0, "V": "0.00000000", "Q": "0.00000000", "B": "0"}'

       # score for this entry would be Open timestamp: 1657651320000

    :param StrictRedis redisClient: A redis connector.
    :param str key: Key redis name.
    :return pd.DataFrame: BinPan dataframe.
    """
    symbol = key.split('@')[0].upper()
    tick_interval = key[-2:]
    data = fetch_zset_range(redisClient=redisClient, key=key.lower(), with_scores=False)
    if not data:
        redis_logger.warning(f"BinPan warning: No data found in Redis for {key}")
    return redis_klines_parser(json_list=data, symbol=symbol, tick_interval=tick_interval)


def zset_length_between_scores(redisClient: StrictRedis,
                               name: str,
                               min_index: int,
                               max_index: int) -> int:
    """
    Returns count of elements between start score and end score (including elements with score equal to min or max).

    :param StrictRedis redisClient: A redis client.
    :param str name: A redis available key (zset name).
    :param min_index: A score presumed value.
    :param max_index: A score presumed value.
    :return int: Count of scores in redis between ()
    """
    return redisClient.zcount(name=name, min=min_index, max=max_index)


def zset_length(redisClient: StrictRedis,
                name: str) -> int:
    """
    Returns count of elements in a set.

    :param StrictRedis redisClient: A redis client.
    :param str name: A redis available key (zset name).
    :return int: Count elements in zset.
    """
    return redisClient.zcard(name=name)


################
# List objects #
################


def push_line_to_redis(redisClient: StrictRedis,
                       key: str, data: str):
    """
    Pushes another line to the end of a list key in redis database.

    :param StrictRedis redisClient: A redis connector.
    :param str key: The redis key name.
    :param str data: Data to push. If not string type will be converted to string.
    :return: redis feedback info.

    """
    return redisClient.rpush(key, str(data))


def fetch_list(redisClient: StrictRedis,
               key: str,
               start_index=0,
               end_index=-1) -> list:
    """
    Fetch list entries in a redis list key from start index to end.

    :param object redisClient: A redis connector.
    :param str key: Key name in redis database.
    :param int start_index: First index to fetch.
    :param int end_index: Last index to fetch.
    :return list: All raw strings in redis list key into a python's list.

    """
    return redisClient.lrange(key, start_index, end_index)


def fetch_data_in_list(redisClient: StrictRedis,
                       key: str,
                       start_index=0,
                       end_index=-1) -> list:
    """
    Fetch data from redis for a given key. Data expected format is redis list.

    .. code-block::

       '{"t": 1657651320000, "o": "0.10100000", "h": "0.10100000", "l": "0.10100000", "c": "0.10100000", "v": "0.00000000",
         "T": 1657651379999, "q": "0.00000000", "n": 0, "V": "0.00000000", "Q": "0.00000000", "B": "0"}'


    :param object redisClient: A redis connector.
    :param str key: An existing redis key.
    :param int start_index: first index in the redis key to fetch for.
    :param int end_index: last index in the redis key to fetch for.
    :return:
    """
    return redisClient.lrange(key, start_index, end_index)


def insert_line_before_index_in_list(redisClient: StrictRedis,
                                     key: str,
                                     idx: int,
                                     value: str):
    """
    Inserts data before an existing string in a redis list. You need to know the index of the reference value.

    :param object redisClient: A redis connector.
    :param str key: A redis key list formatted.
    :param int idx: Index where will be inserted the value, pushing forward existing values, in other words,
        inserting before the referred value in that passed index.
    :param str value: An string to insert.
    :return None:

    """

    idx = int(idx)  # redis does not support numpy types
    old_value = redisClient.lindex(name=key, index=idx)
    tag = 'positional_tag'
    redisClient.lset(name=key, index=idx, value=tag)
    redisClient.linsert(name=key, where='AFTER', refvalue=tag, value=old_value)
    redisClient.linsert(name=key, where='AFTER', refvalue=tag, value=value)
    redisClient.lrem(name=key, count=0, value=tag)


def find_row_index_in_redis_key(redisClient: StrictRedis,
                                symbol: str,
                                tick_interval: str,
                                row_reference: str) -> int:
    """
    It founds in a redis stream the most accurate row with greater timestamp than the passed timestamp row and returns its
    index by iteration.

    :param object redisClient: A redis connector.
    :param str symbol: A binance symbol.
    :param str tick_interval: A binance klines tick interval.
    :param str row_reference: A json.dumps() of a row from binance klines.
    :return int: Index position of the next existing row.
    """
    stream = f"{symbol.lower()}@kline_{tick_interval}"
    my_row_ts = json.loads(row_reference)['t']

    # fetch redis klines full key, got to fetch because it does one by one change
    full_key_list = fetch_list(redisClient=redisClient,
                               key=stream,
                               start_index=0,
                               end_index=-1)
    key_jsons = [json.loads(i) for i in full_key_list]
    df = pd.DataFrame(key_jsons)
    df.loc[:, 'dif'] = df['t'] - my_row_ts
    greater = df[df['dif'] > 0].min()  # expected to get entire row
    if not greater.dropna().empty:
        greater_ts = greater['t']
        next_row = df[df['t'] == greater_ts]
        idx_to_insert = next_row.index[0]
        return idx_to_insert
    else:
        raise Exception("Not contemplated situation!!!!")


def fetch_list_filter_query(redisClient: StrictRedis,
                            coin_filter: str = None,
                            symbol_filter: str = None,
                            tick_interval_filter: str = None,
                            start_index=0,
                            end_index=-1) -> dict:
    """
    Fetch from redis all keys for a time interval. Then filters by symbol or coin in the ticker fetched.

    Is expected to fetch redis list data.

    .. code-block::

       '{"t": 1657651320000, "o": "0.10100000", "h": "0.10100000", "l": "0.10100000", "c": "0.10100000", "v": "0.00000000",
         "T": 1657651379999, "q": "0.00000000", "n": 0, "V": "0.00000000", "Q": "0.00000000", "B": "0"}'

    :param object redisClient: Connector for redis.
    :param str coin_filter: A coin string to filter out all redis keys without it.
    :param str symbol_filter: Optional. A symbol string to filter out all redis keys without it. Symbol filter prevales over coin filter.
    :param str tick_interval_filter: Binance klines tick interval.
    :param int start_index: first index in the redis key to fetch for.
    :param int end_index: last index in the redis key to fetch for.
    :return dict: A dict each symbol with its data.

    """

    if symbol_filter:
        symbol_filter = symbol_filter.lower()
    if coin_filter:
        coin_filter = coin_filter.lower()

    all_keys = fetch_keys(redisClient=redisClient,
                          filter_tick_interval=tick_interval_filter)
    if symbol_filter:
        all_keys = [i for i in all_keys if symbol_filter in i]
    elif coin_filter:
        all_keys = [i for i in all_keys if coin_filter in i]

    ret = {}
    for ticker in all_keys:
        data = fetch_data_in_list(redisClient=redisClient,
                                  key=ticker,
                                  start_index=start_index,
                                  end_index=end_index)
        ret[ticker] = data
    return ret


#############
# pipelines #
#############

# TODO: actualizar con lo nuevo de binance cache

def execute_pipeline(pipeline: StrictRedis.pipeline):
    """
    Executes a pipeline buffer.

    :param redisClient pipeline: A redis client pipeline.
    :return: Redis response.
    """
    return pipeline.execute()


def flush_pipeline(pipeline: StrictRedis.pipeline) -> StrictRedis.pipeline:
    """
    Flush pipeline buffered commands.

    :param redisClient pipeline: A redis client pipeline.
    :return: Redis response.
    """
    return pipeline.command_stack


def pipe_buffer_ordered_set(pipeline: StrictRedis,
                            key: str,
                            mapping: dict,
                            LT=False,
                            XX=False,
                            NX=True,
                            GT=False,
                            CH=True,
                            INCR=False,
                            ):
    """
        Fills a pipeline buffer to "pipeline.execute()" later with commands of pushes of elements to an ordered set with a score as index.

        :param redisClient pipeline: A redis client pipeline.
        :param str key: The redis key name.
        :param list mapping: Data to push in the format {data: score}
        :param LT: Only update existing elements if the new score is less than the current score. This flag doesn't prevent adding new elements.
        :param XX: Only update elements that already exist. Don't add new elements.
        :param NX: Only add new elements. Don't update already existing elements. This is the default only true value.
        :param GT: Only update existing elements if the new score is greater than the current score. This flag doesn't prevent adding new
           elements.
        :param CH: Modify the return value from the number of new elements added, to the total number of elements changed (CH is an abbreviation
           of changed). Changed elements are new elements added and elements already existing for which the score was updated. So elements
           specified in the command line having the same score as they had in the past are not counted. Note: normally the return value of ZADD
           only counts the number of new elements added.
        :param INCR: When this option is specified ZADD acts like ZINCRBY. Only one score-element pair can be specified in this mode.
           Note: The GT, LT and NX options are mutually exclusive.
        :return: redis feedback info.

    """
    pipeline.zadd(name=key, mapping=mapping, lt=LT, xx=XX, nx=NX, gt=GT, ch=CH, incr=INCR)
    return pipeline


def pipe_zset_range(pipeline: StrictRedis.pipeline,
                    key: str,
                    start_index=0,
                    end_index=-1,
                    with_scores=False) -> StrictRedis.pipeline:
    """
    Pipe a request for an ordered set by index for a redis existing key.

    :param StrictRedis.pipeline pipeline: Pipeline object.
    :param str key: Key to pipe wanted data.
    :param int start_index: Starting index of data. If ordered by scoring timestamps, starting values are older than ending values.
        Default 0
    :param int end_index: Ending index of data. If ordered by scoring timestamps, starting values are older than ending values.
        Default -1
    :param bool with_scores: Pipes for a list of tuples with data and score.
    :return StrictRedis.pipeline: Pipeline object.
    """
    return pipeline.zrange(name=key, start=start_index, end=end_index, withscores=with_scores)


def pipe_zset_timestamps(pipeline: StrictRedis.pipeline,
                         key: str,
                         start_timestamp: int,
                         end_timestamp: int = None,
                         with_scores=False) -> StrictRedis.pipeline:
    """
    Puts into pipeline buffer a redis ordered set by its key name and start end scores. Optionally with redis index (scores)
    in a tuple each row. Optionally returns with redis index (scores) in a tuple each row if with_Scores.

    :param StrictRedis.pipeline pipeline: A redis pipeline.
    :param str key: Redis key name.
    :param int start_timestamp: Usually a timestamp as index in redis key.
    :param int end_timestamp: Usually a timestamp as index in redis key.
    :param bool with_scores: Will return rows with its index in a tuple if true.
    :return StrictRedis.pipeline: A pipeline ready to execute.
    """
    return pipeline.zrangebyscore(name=key,
                                  min=start_timestamp,
                                  max=end_timestamp,
                                  withscores=with_scores)


def pipe_time_interval_bulk_ohlc_data(pipeline: StrictRedis.pipeline,
                                      min_index: int,
                                      max_index: int,
                                      my_keys: list,
                                      length=202,
                                      batch=200) -> dict:
    """
    Pipes to redis, check time integrity and parse in a list of tuples with stream and open, high, low, close arrays.
    """

    ohlc_arrays = {}

    for i in range(0, len(my_keys), batch):
        batch_keys = my_keys[i:i + batch]
        for key in batch_keys:
            pipeline = pipe_zset_timestamps(pipeline=pipeline,
                                            key=key,
                                            start_timestamp=min_index,
                                            end_timestamp=max_index)

        raw_klines = pipeline.execute()
        redis_logger.debug(f"Data to parse: {bool(any(raw_klines))} not empty len:{len([n for n in raw_klines if n])}")
        if not any(raw_klines):
            continue

        # check length
        for ii, k in enumerate(batch_keys):
            if len(raw_klines[ii]) != length:
                redis_logger.info(f"Missing klines for {k}")
                continue
            tick_interval = k.split('_')[-1]
            if klines_continuity(klines=raw_klines[ii], tick_interval=tick_interval):
                # parseo en numpy OHLC
                ohlc_arrays[k] = klines_ohlc_to_numpy(klines=raw_klines[ii])
            else:
                redis_logger.info(f"No continuity for {k}")

    if my_keys:
        redis_logger.debug(f"From {len(my_keys)} keys, valid: {len(ohlc_arrays)} - pct: {100 * len(ohlc_arrays) / len(my_keys):.2f}")
    else:
        redis_logger.info("No keys fetched!!!")
    return ohlc_arrays



import pandas as pd
from random import sample
from .logs import Logs


tags_logger = Logs(filename='./logs/tags_logger.log', name='tags_logger', info_level='INFO')


def random_strategy(data: pd.DataFrame,
                    buys_qty: int = 10,
                    sells_qty: int = 10,
                    new_actions_col: str = 'actions',
                    labels: dict = {'buy': 1, 'sell': -1},
                    fill: str = None):
    """
    Creates a random buy and sell tag in a column of a dataframe.

    :param pd.DataFrame data: A dataframe.
    :param int buys_qty: Buy tags quantity. Default is 10.
    :param int sells_qty: Sells tags quantity. Default is 10.
    :param str new_actions_col: Name of the new column with the buy and sell tags. Defaults to 'actions'
    :param dict labels: Tags for each action. Default is {'buy': 1, 'sell': -1}.
    :param str fill: Optional, fills nans with some string.
    :return: pd.DataFrame

    """
    df = data.copy(deep=True)
    buys = sample(list(df.index), buys_qty)
    sells = sample(list(df.index), sells_qty)
    df.loc[buys, new_actions_col] = labels['buy']
    df.loc[sells, new_actions_col] = labels['sell']
    if fill:
        df[new_actions_col].fillna(fill, inplace=True)
    return df



"""

Market functions

"""

from tqdm import tqdm
from time import time
import pandas as pd
import json
from decimal import Decimal as dd
from redis import StrictRedis
from typing import List

import handlers.time_helper
from .logs import Logs
from .quest import check_weight, get_response, api_raw_get, get_semi_signed_request
from .time_helper import tick_seconds, end_time_from_start_time, start_time_from_end_time
from secret import api_key

market_logger = Logs(filename='./logs/market_logger.log', name='market_logger', info_level='INFO')

base_url = 'https://api.binance.com'

klines_columns = {"t": "Open time",
                  "o": "Open",
                  "h": "High",
                  "l": "Low",
                  "c": "Close",
                  "v": "Volume",
                  "T": "Close time",
                  "q": "Quote volume",
                  "n": "Trades",
                  "V": "Taker buy base volume",
                  "Q": "Taker buy quote volume",
                  "B": "Ignore"}

trades_columns = {'M': 'Best price match',
                  'm': 'Buyer was maker',
                  'T': 'Timestamp',
                  'l': 'Last tradeId',
                  'f': 'First tradeId',
                  'q': 'Quantity',
                  'p': 'Price',
                  'a': 'Aggregate tradeId'}


##########
# Prices #
##########


def get_last_price(symbol: str = None) -> dict or list:
    """
    Returns all prices of symbols in a dict or if symbol specified, the float price of symbol.

    :param str symbol: A binance symbol.
    :return dict or list:

    """
    endpoint = '/api/v3/ticker/price'
    if symbol:
        weight = 1
        symbol = symbol.upper()
    else:
        weight = 2
    res = api_raw_get(endpoint=endpoint, weight=weight, params={'symbol': symbol})
    market_logger.debug(f"get_last_price: {res}")

    if type(res) == dict:
        return float(res['price'])
    elif type(res) == list:
        return {k['symbol']: float(k['price']) for k in res}


def get_prices_dic(decimal_mode: bool) -> dict:
    """
    Gets all symbols current prices into a dictionary.
    :decimal_mode bool: It flags to work in decimal mode.
    :return dict:
    """
    endpoint = '/api/v3/ticker/price'
    check_weight(2, endpoint=endpoint)
    ret = get_response(url=endpoint)
    if decimal_mode:
        return {d['symbol']: dd(d['price']) for d in ret}
    else:
        return {d['symbol']: float(d['price']) for d in ret}


###########
# Candles #
###########

def get_candles_by_time_stamps(symbol: str,
                               tick_interval: str,
                               start_time: int = None,
                               end_time: int = None,
                               limit=1000,
                               time_zone='Europe/Madrid',
                               redis_client: StrictRedis or dict = None) -> list:
    """
    Calls API for a candles list using one or two timestamps, starting and ending.

    In case the limit is passed and exceeded by requested time intervals, the start_time prevails over the end_time,
    start_time must come in milliseconds from epoch.

    In case of two timeStamps as arguments, limit is ignored.

    The API rounds the startTime up to the next open of the next candle. That is, it does not include the candle in which there is
    that timeStamp, but the next candle of the corresponding tick_interval, except in case it exactly matches the value of an open
    timestamp, in which case it will include it in the return.

    The indicated endTime will include the candlestick that timestamp is on. It should be in milliseconds. It can be a not closed one if
    is open right in between the endtime timestamp.

    If no timestamps are passed, the last quantity candlesticks up to limit count are returned.

    :param str symbol: A binance valid symbol.
    :param str tick_interval: A binance valid time interval for candlesticks.
    :param int start_time: A timestamp in milliseconds from epoch.
    :param int end_time: A timestamp in milliseconds from epoch.
    :param int limit: Count of candles to ask for.
    :param bool or StrictRedis or dict redis_client: A redis instance of a connector. Also can be passed a dictionary with redis client configuration. Example:

        redis_client = {'host': '192.168.89.242', 'port': 6379, 'db': 0, 'decode_responses': True}

    :param str time_zone: Just used for exception errors.
    :return list: Returns a list from the Binance API

    .. code-block::

        [
          [
            1499040000000,      // Open time
            "0.01634790",       // Open
            "0.80000000",       // High
            "0.01575800",       // Low
            "0.01577100",       // Close
            "148976.11427815",  // Volume
            1499644799999,      // Close time
            "2434.19055334",    // Quote asset volume
            308,                // Number of trades
            "1756.87402397",    // Taker buy base asset volume
            "28.46694368",      // Taker buy quote asset volume
            "17928899.62484339" // Ignore.
          ]
        ]

    """
    endpoint = '/api/v3/klines?'

    start_string = handlers.time_helper.convert_milliseconds_to_str(ms=start_time, timezoned=time_zone)
    end_string = handlers.time_helper.convert_milliseconds_to_str(ms=end_time, timezoned=time_zone)

    market_logger.debug(
        f"get_candles_by_time_stamps -> symbol={symbol} tick_interval={tick_interval} start={start_string} end={end_string}")

    tick_milliseconds = int(tick_seconds[tick_interval] * 1000)

    if not start_time and not end_time:
        end_time = int(time()) * 1000

    if not start_time and end_time:
        start_time = end_time - (limit * tick_milliseconds)
        start_time = handlers.time_helper.open_from_milliseconds(ms=start_time, tick_interval=tick_interval)

    elif start_time and not end_time:
        end_time = start_time + (limit * tick_milliseconds)  # ??? for getting limit exactly
        end_time = handlers.time_helper.open_from_milliseconds(ms=end_time, tick_interval=tick_interval)

    # prepare iteration for big loops
    tick_milliseconds = int(tick_seconds[tick_interval] * 1000)
    ranges = [(i, i + (1000 * tick_milliseconds)) for i in range(start_time, end_time, tick_milliseconds * 1000)]

    # loop
    raw_candles = []
    for r in ranges:
        start = r[0]
        end = r[1]

        if redis_client:
            if type(redis_client) == dict:
                try:
                    redis_client = StrictRedis(**redis_client)
                except Exception:
                    msg = f"BinPan exceptionRedis client error: arguments for client={redis_client}"
                    market_logger.error(msg)
                    raise Exception(msg)
            elif type(redis_client) != StrictRedis:
                msg = f"BinPan exceptionRedis client error: type passed={type(redis_client)}"
                market_logger.error(msg)
                raise Exception(msg)

            ret = redis_client.zrangebyscore(name=f"{symbol.lower()}@kline_{tick_interval}",
                                             min=start,
                                             max=end,
                                             withscores=False)
            if not ret:
                continue
            response = [json.loads(i) for i in ret]
        else:
            params = {'symbol': symbol,
                      'interval': tick_interval,
                      'startTime': start,
                      'endTime': end,
                      'limit': limit}
            params = {k: v for k, v in params.items() if v}
            check_weight(1, endpoint=endpoint)
            response = get_response(url=endpoint, params=params)

        raw_candles += response

    if not raw_candles:
        msg = f"BinPan Exception: Requested data for {symbol.lower()}@kline_{tick_interval} between {start_string} and " \
              f"{end_string} missing in redis server: ."
        market_logger.error(msg)
        raise Exception(msg)

    # descarta sobrantes
    overtime_candle_ts = handlers.time_helper.next_open_by_milliseconds(ms=end_time, tick_interval=tick_interval)
    if raw_candles:
        if type(raw_candles[0]) == list:  # if from binance
            raw_candles = [i for i in raw_candles if int(i[0]) < overtime_candle_ts]
        else:
            open_ts_key = list(raw_candles[0].keys())[0]
            raw_candles = [i for i in raw_candles if int(i[open_ts_key]) < overtime_candle_ts]

    return raw_candles


def parse_candles_to_dataframe(raw_response: list,
                               symbol: str,
                               tick_interval: str,
                               columns: list = None,
                               time_cols: list = None,
                               time_zone: str or None = 'UTC',
                               time_index=False) -> pd.DataFrame:
    """
    Format a list of lists by changing the indicated time fields to string format.

    Passing a time_zone, for example 'Europe/Madrid', will change the time from utc to the indicated zone.

    It will automatically sort the DataFrame using the first column of the time_cols list.

    The index of the DataFrame will be numeric correlative.

    :param list(lists) raw_response:        API klines response. List of lists.
    :param str symbol:          Symbol requested
    :param str tick_interval:   Tick interval between candles.
    :param list columns:         Column names. Default is BinPan dataframe columns.
    :param list time_cols:       Columns to take dates from.
    :param str or None time_zone: Optional. Time zone to convert dates in index.
    :param bool time_index:      True gets dates index, False just numeric index.
    :return:                    Pandas DataFrame

    """
    if not time_cols:
        time_cols = ['Open time', 'Close time']

    if not columns:
        columns = ['Open time', 'Open', 'High', 'Low', 'Close', 'Volume', 'Close time', 'Quote volume',
                   'Trades', 'Taker buy base volume', 'Taker buy quote volume', 'Ignore']

    columns = [col for col in columns if not 'timestamp' in col]

    # check if redis columns
    if raw_response:
        if type(raw_response[0]) == list:
            df = pd.DataFrame(raw_response, columns=columns)
        else:
            response_keys = list(raw_response[0].keys())
            response_keys.sort()
            sort_columns = columns
            sort_columns.sort()

            if response_keys != sort_columns:  # json keys from redis different
                columns = list(klines_columns.keys())
                df = pd.DataFrame(raw_response, columns=columns)
                df.rename(columns=klines_columns, inplace=True)
            else:
                df = pd.DataFrame(raw_response, columns=columns)
    else:
        msg = f"BinPan Warning: No response to parse for {symbol} {tick_interval}. " \
              f"Check symbol requested time interval or {symbol} online status: binpan.Exchange().df.loc['{symbol}'].to_dict()['status']"
        market_logger.warning(msg)
        return pd.DataFrame(columns=columns + ['Open timestamp', 'Close timestamp'])

    # for col in df.columns:
    #     df[col] = pd.to_numeric(arg=df[col], downcast='integer')
    df = convert_to_numeric(data=df)

    df.loc[:, 'Open timestamp'] = df['Open time']
    df.loc[:, 'Close timestamp'] = df['Close time']

    if type(time_zone) == str and time_zone != 'UTC':  # converts to time zone the time columns
        for col in time_cols:
            df.loc[:, col] = handlers.time_helper.convert_utc_ms_column_to_time_zone(df, col, time_zone=time_zone)
            df.loc[:, col] = df[col].apply(lambda x: handlers.time_helper.convert_datetime_to_string(x))
    else:
        for col in time_cols:
            df.loc[:, col] = df[col].apply(lambda x: handlers.time_helper.convert_milliseconds_to_utc_string(x))

    if time_index and time_zone:
        date_index = df['Open timestamp'].apply(handlers.time_helper.convert_milliseconds_to_time_zone_datetime, timezoned=time_zone)
        df.set_index(date_index, inplace=True)
        index_name = f"{symbol} {tick_interval} {time_zone}"
        df.index.name = index_name
    else:
        index_name = f"{symbol} {tick_interval}"
        df.index.name = index_name

    # if came from a loop there are duplicated values in step connections to be removed.
    df.drop_duplicates(inplace=True)
    return df


###############
# format data #
###############

def basic_dataframe(data: pd.DataFrame,
                    exceptions: list = None,
                    actions_col='actions'
                    ) -> pd.DataFrame:
    """
    Delete all columns except: Open, High, Low, Close, Volume, actions.

    Some columns can be excepted.

    Useful to drop messed technical indicators columns in one shot.

    :param pd.DataFrame data:        A BinPan DataFrame
    :param list exceptions:  A list of columns to avoid dropping.
    :param str actions_col: A column with operative actions.
    :return pd.DataFrame: Pandas DataFrame

    """
    df_ = data.copy(deep=True)
    if actions_col not in df_.columns:
        if exceptions:
            return df_[['Open', 'High', 'Low', 'Close', 'Volume'] + exceptions].copy(deep=True)
        else:
            return df_[['Open', 'High', 'Low', 'Close', 'Volume']].copy(deep=True)
    else:
        if exceptions:
            return df_[['Open', 'High', 'Low', 'Close', 'Volume', actions_col] + exceptions].copy(deep=True)
        else:
            return df_[['Open', 'High', 'Low', 'Close', 'Volume', actions_col]].copy(deep=True)


def convert_to_numeric(data: pd.DataFrame) -> pd.DataFrame:
    """
    Converts to numeric all posible columns for a given Dataframe.

    :param pd.DataFrame data: A dataframe with columns
    :return pd.DataFrame: A dataframe with numeric values in each column that can be numeric.
    """

    for col in data.columns:
        try:
            data.loc[:, col] = pd.to_numeric(arg=data[col], downcast='integer', errors='raise')
        except Exception:
            pass
    return data


##########
# Trades #
##########


def get_agg_trades(symbol: str,
                   fromId: int = None,
                   toId: int = None,
                   limit=None,
                   startTime: int = None,
                   endTime: int = None,
                   redis_client_trades: StrictRedis = None) -> List[dict]:
    """
    Returns aggregated trades from id to limit or last trades if id not specified. Also is possible to get from starTime utc in
    milliseconds from epoch or until endtime milliseconds from epoch.

    If it is tested with more than 1 hour of trades, it gives error 1127 and if you adjust it to one hour,
    the maximum limit of 1000 is NOT applied.

    Start time and end time not applied if trade id passed.

    Limit applied in fromId mode defaults to 500. Maximum is 1000.

    :param str symbol: A binance valid symbol.
    :param int fromId: An aggregated trade id.
    :param int toId: An aggregated trade id end, used just with redis.
    :param int limit: Count of trades to ask for.
    :param int startTime: A timestamp in milliseconds from epoch.
    :param int endTime: A timestamp in milliseconds from epoch.
    :param bool redis_client_trades: A redis instance of a connector. Must be a trades redis connector, usually different configuration
     from candles redis server.
    :return list: Returns a list from the Binance API in dicts.

    .. code-block::

        [
          {
            "a": 26129,         // Aggregate tradeId
            "p": "0.01633102",  // Price
            "q": "4.70443515",  // Quantity
            "f": 27781,         // First tradeId
            "l": 27781,         // Last tradeId
            "T": 1498793709153, // Timestamp
            "m": true,          // Was the buyer the maker?
            "M": true           // Was the trade the best price match?
          }
        ]
    """

    endpoint = '/api/v3/aggTrades?'
    check_weight(1, endpoint=endpoint)

    if redis_client_trades:
        if fromId and limit:

            response = redis_client_trades.zrangebyscore(name=f"{symbol.lower()}@aggTrade",
                                                         min=fromId,
                                                         max=fromId + limit,
                                                         withscores=False)
        elif fromId and toId:
            response = redis_client_trades.zrangebyscore(name=f"{symbol.lower()}@aggTrade",
                                                         min=fromId,
                                                         max=toId,
                                                         withscores=False)

        elif startTime and endTime:

            response = []
            market_logger.info(f"Fetching all trades from redis server for {symbol}")

            curr_trades = redis_client_trades.zrange(name=f"{symbol.lower()}@aggTrade",
                                                     start=0,
                                                     end=-1,
                                                     withscores=True)

            curr_trades = [json.loads(i[0]) for i in curr_trades]
            market_logger.debug(f"Fetched {len(curr_trades)} aggregate trades for {symbol}")

            if curr_trades:
                timestamps_dict = {k['T']: k['a'] for k in curr_trades}
                trade_ids = [i for t, i in timestamps_dict.items() if startTime <= t <= endTime]
                market_logger.debug(f"List of trades {len(trade_ids)} for {symbol}")

                if trade_ids:
                    response = redis_client_trades.zrangebyscore(name=f"{symbol.lower()}@aggTrade",
                                                                 min=trade_ids[0],
                                                                 max=trade_ids[-1],
                                                                 withscores=False)
                    response = [json.loads(i) for i in response]
                else:
                    response = []

                market_logger.info(f"Clean aggregated {len(response)} trades found for {symbol}")

            if not response:
                start_str = handlers.time_helper.convert_milliseconds_to_utc_string(ms=startTime)
                end_str = handlers.time_helper.convert_milliseconds_to_utc_string(ms=endTime)
                market_logger.info(f"No trade IDs found for {symbol} for given interval {start_str} and {end_str} (UTC) in server.")

        else:
            market_logger.info(f"Request for trades from {symbol} returning ALL trades available in Redis.")

            ret_raw = redis_client_trades.zrange(name=f"{symbol.lower()}@aggTrade",
                                                 start=0,
                                                 end=-1,
                                                 withscores=False)
            response = []
            pbar = tqdm(ret_raw)
            for i in pbar:
                pbar.set_description(desc=f"Aggregate Trades for {symbol}", refresh=True)
                response.append(json.loads(i))

    else:
        if fromId and not startTime and not endTime:
            query = {'symbol': symbol, 'limit': limit, 'fromId': fromId}

        elif startTime and endTime:  # Limited to one hour by api

            try:
                assert (endTime - startTime) <= (1000 * 60 * 60)
            except AssertionError:
                msg = f"BinPan Error: Aggregate trades timestamps interval cannot be greater than 1 hour. Please use historical aggregate " \
                      f"trades function: interval in minutes = {(endTime - startTime) / (1000 * 60)} start: {startTime} end: {endTime}"
                market_logger.error(msg)
                raise Exception(msg)

            query = {'symbol': symbol, 'limit': limit, 'startTime': startTime, 'endTime': endTime}

        elif startTime or endTime:
            # Limited to one hour by api
            query = {'symbol': symbol, 'limit': limit}
            if startTime and not endTime:
                query.update({'startTime': startTime})
                query.update({'endTime': end_time_from_start_time(startTime=startTime,
                                                                  limit=1,
                                                                  tick_interval='1h')})
            if endTime and not startTime:
                query.update({'endTime': endTime})
                query.update({'startTime': start_time_from_end_time(endTime=endTime,
                                                                    limit=1,
                                                                    tick_interval='1h')})
        else:  # last ones
            query = {'symbol': symbol, 'limit': limit}

        response = get_response(url=endpoint, params=query)

    return response


def get_historical_aggregated_trades(symbol: str,
                                     startTime: int,
                                     endTime: int,
                                     redis_client_trades: StrictRedis = None) -> List[dict]:
    """
    Returns aggregated trades between timestamps. It iterates over 1 hour intervals to avoid API one hour limit.

    :param int startTime: A timestamp in milliseconds from epoch.
    :param int endTime: A timestamp in milliseconds from epoch.
    :param str symbol: A binance valid symbol.
    :param bool redis_client_trades: A redis instance of a connector. Must be a trades redis connector, usually different configuration
     from candles redis server.
    :return list: Returns a list from the Binance API

    .. code-block::

        [
          {
            "a": 26129,         // Aggregate tradeId
            "p": "0.01633102",  // Price
            "q": "4.70443515",  // Quantity
            "f": 27781,         // First tradeId
            "l": 27781,         // Last tradeId
            "T": 1498793709153, // Timestamp
            "m": true,          // Was the buyer the maker?
            "M": true           // Was the trade the best price match?
          }
        ]
    """
    hour_ms = 60 * 60 * 1000
    if (endTime - startTime) <= hour_ms:
        return get_agg_trades(symbol=symbol, startTime=startTime, endTime=endTime, redis_client_trades=redis_client_trades)
    else:
        response = []
        if not redis_client_trades:
            pbar = tqdm(range(startTime, endTime, hour_ms))
            for i in pbar:
                pbar.set_description(desc=f"Historical Aggregate Trades for {symbol}", refresh=True)
                response += get_agg_trades(symbol=symbol, startTime=i, endTime=i + hour_ms, redis_client_trades=redis_client_trades)
        else:
            response = []
            market_logger.info(f"Fetching all trades from redis server for {symbol}")

            curr_trades = redis_client_trades.zrange(name=f"{symbol.lower()}@aggTrade",
                                                     start=0,
                                                     end=-1,
                                                     withscores=True)

            curr_trades = [json.loads(i[0]) for i in curr_trades]
            market_logger.debug(f"Fetched {len(curr_trades)} aggregate trades for {symbol}")

            if curr_trades:
                timestamps_dict = {k['T']: k['a'] for k in curr_trades}
                trade_ids = [i for t, i in timestamps_dict.items() if startTime <= t <= endTime]
                market_logger.debug(f"List of trades {len(trade_ids)} for {symbol}")

                if trade_ids:
                    response = redis_client_trades.zrangebyscore(name=f"{symbol.lower()}@aggTrade",
                                                                 min=trade_ids[0],
                                                                 max=trade_ids[-1],
                                                                 withscores=False)
                    response = [json.loads(i) for i in response]
                else:
                    response = []

                market_logger.info(f"Clean aggregated {len(response)} trades found for {symbol}")

            if not response:
                start_str = handlers.time_helper.convert_milliseconds_to_utc_string(ms=startTime)
                end_str = handlers.time_helper.convert_milliseconds_to_utc_string(ms=endTime)
                market_logger.info(f"No trade IDs found for {symbol} for given interval {start_str} and {end_str} (UTC) in server.")
        return response


def parse_agg_trades_to_dataframe(response: list,
                                  columns: dict,
                                  symbol: str,
                                  time_zone: str = None,
                                  time_index: bool = None):
    """
    Parses the API response into a pandas dataframe.

    .. code-block::

         {'M': True,
          'T': 1656166914571,
          'a': 1218761712,
          'f': 1424997754,
          'l': 1424997754,
          'm': True,
          'p': '21185.05000000',
          'q': '0.03395000'}

    :param list response: API raw response from trades.
    :param columns: Column names.
    :param symbol: The used symbol.
    :param time_zone: Selected time zone.
    :param time_index: Or integer index.
    :return: Pandas DataFrame

    """
    if not response:
        return pd.DataFrame(columns=list(columns.values()))

    df = pd.DataFrame(response)
    df.rename(columns=columns, inplace=True)
    # df.loc[:, 'Buyer was maker'] = df['Buyer was maker'].replace({'Maker buyer': 1, 'Taker buyer': 0})

    df = convert_to_numeric(data=df)

    timestamps_col = 'Timestamp'
    timestamps_serie = df[timestamps_col].copy()
    col = 'Date'
    df.loc[:, col] = df[timestamps_col]
    if time_zone != 'UTC':  # converts to time zone the time columns
        df.loc[:, col] = handlers.time_helper.convert_utc_ms_column_to_time_zone(df, col, time_zone=time_zone)
        df.loc[:, col] = df[col].apply(lambda x: handlers.time_helper.convert_datetime_to_string(x))
    else:
        df.loc[:, col] = df[timestamps_col].apply(lambda x: handlers.time_helper.convert_milliseconds_to_utc_string(x))

    if time_index:
        date_index = timestamps_serie.apply(handlers.time_helper.convert_milliseconds_to_time_zone_datetime, timezoned=time_zone)
        df.set_index(date_index, inplace=True)

    index_name = f"{symbol} {time_zone}"
    df.index.name = index_name
    # df.loc[:, 'Buyer was maker'] = df['Buyer was maker'].astype(bool)
    return df[['Aggregate tradeId', 'Price', 'Quantity', 'First tradeId', 'Last tradeId', 'Date', 'Timestamp', 'Buyer was maker',
               'Best price match']]


def get_last_atomic_trades(symbol: str,
                           limit=1000):
    """
    Returns recent atomic (not aggregated) trades.

    GET /api/v3/trades

    Get recent trades.

    Weight(IP): 1

    :param str symbol: A binance valid symbol.
    :param int limit: API max limit is 1000.
    :return list: Returns a list from the Binance API

    Return example:

    .. code-block::

        [
          {
            "id": 28457,
            "price": "4.00000100",
            "qty": "12.00000000",
            "quoteQty": "48.000012",
            "time": 1499865549590,
            "isBuyerMaker": true,
            "isBestMatch": true
          }
        ]

    """

    endpoint = '/api/v3/trades?'
    check_weight(1, endpoint=endpoint)
    query = {'symbol': symbol, 'limit': limit}
    return get_response(url=endpoint, params=query)


def get_atomic_trades(symbol: str,
                      fromId: int = None,
                      limit: int = None,
                      decimal_mode: bool = False) -> list:
    """
    Returns atomic (not aggregated) trades from id to limit or last trades if id not specified.

    Limit applied in fromId mode defaults to 500. Maximum is 1000.

    GET /api/v3/historicalTrades

    Get older market trades.

    Weight(IP): 5

    :param int fromId: Trade id to fetch from. If not passed, gets most recent trades.
    :param str symbol: A binance valid symbol.
    :param int limit: Count of trades to ask for.
    :param bool decimal_mode: Enables decimal type for return.
    :return list: Returns a list from the Binance API.

    .. code-block::

        [
          {
            "id": 28457,
            "price": "4.00000100",
            "qty": "12.00000000",
            "quoteQty": "48.000012",
            "time": 1499865549590, // Trade executed timestamp, as same as `T` in the stream
            "isBuyerMaker": true,
            "isBestMatch": true
          }
        ]
    """

    endpoint = '/api/v3/historicalTrades?'
    check_weight(5, endpoint=endpoint)
    query = {'symbol': symbol, 'limit': limit, 'fromId': fromId, 'recWindow': None}
    return get_semi_signed_request(url=endpoint,
                                   decimal_mode=decimal_mode,
                                   api_key=api_key,
                                   params=query)


def get_historical_atomic_trades(symbol: str,
                                 startTime: int = None,
                                 endTime: int = None,
                                 start_trade_id: int = None,
                                 end_trade_id: int = None,
                                 limit: int = 1000,
                                 redis_client_trades: StrictRedis = None) -> List[dict]:
    """
    Returns atomic (not aggregated) trades between timestamps. It iterates over limit 1000 intervals to adjust to API limit.

    This request can be very slow because the API request weight limit.

    :param int startTime: A timestamp in milliseconds from epoch.
    :param int endTime: A timestamp in milliseconds from epoch.
    :param int start_trade_id: A trade id as first one (older).
    :param int end_trade_id: A trade id as last one (newer).
    :param int limit: Limit for missing heads or tails of the interval requested with timestamps or trade ids.
    :param str symbol: A binance valid symbol.
    :param bool redis_client_trades: A redis instance of a connector. Must be a trades redis connector, usually different configuration
     from candles redis server.
    :return list: Returns a list from the Binance API

    .. code-block::

        Last Trades example:
            [
                {'id': 86206215,
                 'price': '0.00454100',
                 'qty': '0.02400000',
                 'quoteQty': '0.00010898',
                 'time': 1669579405932,
                 'isBuyerMaker': False,
                 'isBestMatch': True}, ...
             ]
    """

    if start_trade_id and not end_trade_id:
        end_trade_id = start_trade_id + limit
    elif end_trade_id and not start_trade_id:
        start_trade_id = end_trade_id - limit

    trades = handlers.market.get_last_atomic_trades(symbol=symbol, limit=1000)
    requests_cnt = 0
    # with trade ids
    if start_trade_id and end_trade_id and not redis_client_trades:
        current_first_trade = trades[0]['id']
        while trades[0]['id'] > start_trade_id:
            requests_cnt += 1
            market_logger.info(f"Requests to API for atomic trades of {symbol}: {requests_cnt}")
            fetched_older_trades = handlers.market.get_atomic_trades(symbol=symbol,
                                                                     fromId=(current_first_trade - 1000),
                                                                     limit=1000)
            trades = fetched_older_trades + trades
            current_first_trade = trades[0]['id']
        return [i for i in trades if start_trade_id <= i['id'] <= end_trade_id]

    # with timestamps
    if not redis_client_trades:
        current_first_trade = trades[0]['id']

        while trades[0]['time'] > startTime:
            requests_cnt += 1
            market_logger.info(f"Requests to API for atomic trades of {symbol}: {requests_cnt}")
            fetched_older_trades = handlers.market.get_atomic_trades(symbol=symbol,
                                                                     fromId=(current_first_trade - 1000),
                                                                     limit=1000)
            trades = fetched_older_trades + trades
            current_first_trade = trades[0]['id']
        return [i for i in trades if startTime <= i['time'] <= endTime]

    # from redis
    response = []
    market_logger.info(f"Fetching atomic trades from redis server for {symbol}")

    curr_trades = redis_client_trades.zrange(name=f"{symbol.lower()}@trade",
                                             start=0,
                                             end=-1,
                                             withscores=True)

    curr_trades = [json.loads(i[0]) for i in curr_trades]
    market_logger.debug(f"Fetched {len(curr_trades)} atomic trades for {symbol}")

    if curr_trades:
        timestamps_dict = {k['T']: k['t'] for k in curr_trades}
        trade_ids = [i for t, i in timestamps_dict.items() if startTime <= t <= endTime]
        market_logger.debug(f"List of trades {len(trade_ids)} for {symbol}")

        if trade_ids:
            response = redis_client_trades.zrangebyscore(name=f"{symbol.lower()}@trade",
                                                         min=trade_ids[0],
                                                         max=trade_ids[-1],
                                                         withscores=False)
            response = [json.loads(i) for i in response]
        else:
            response = []

        market_logger.info(f"Clean atomic {len(response)} trades found for {symbol}")

    if not response:
        start_str = handlers.time_helper.convert_milliseconds_to_utc_string(ms=startTime)
        end_str = handlers.time_helper.convert_milliseconds_to_utc_string(ms=endTime)
        market_logger.info(f"No atomic trade IDs found for {symbol} for given interval {start_str} and {end_str} (UTC) in server.")
    return response


def parse_atomic_trades_to_dataframe(response: list,
                                     columns: dict,
                                     symbol: str,
                                     time_zone: str = None,
                                     time_index: bool = None):
    """
    Parses the API response into a pandas dataframe.

    .. code-block::

        [
          {
            "id": 28457,
            "price": "4.00000100",
            "qty": "12.00000000",
            "quoteQty": "48.000012",
            "time": 1499865549590, // Trade executed timestamp, as same as `T` in the stream
            "isBuyerMaker": true,
            "isBestMatch": true
          }
        ]

    Redis response:

     {'e': 'trade',
      'E': 1669721069347,
      's': 'LTCBTC',
      't': 86293174,
      'p': '0.00466300',
      'q': '3.10900000',
      'b': 992385453,
      'a': 992385448,
      'T': 1669721069347,
      'm': False,
      'M': True},
     ...]

    :param list response: API raw response from atomic trades.
    :param columns: Column names.
    :param symbol: The used symbol.
    :param time_zone: Selected time zone.
    :param time_index: Or integer index.
    :return: Pandas DataFrame

    """
    if not response:
        return pd.DataFrame(columns=list(columns.values()))

    df = pd.DataFrame(response)
    df.rename(columns=columns, inplace=True)

    df = convert_to_numeric(data=df)

    timestamps_col = 'Timestamp'
    timestamps_serie = df[timestamps_col].copy()
    col = 'Date'
    df.loc[:, col] = df[timestamps_col]
    if time_zone != 'UTC':  # converts to time zone the time columns
        df.loc[:, col] = handlers.time_helper.convert_utc_ms_column_to_time_zone(df, col, time_zone=time_zone)
        df.loc[:, col] = df[col].apply(lambda x: handlers.time_helper.convert_datetime_to_string(x))
    else:
        df.loc[:, col] = df[timestamps_col].apply(lambda x: handlers.time_helper.convert_milliseconds_to_utc_string(x))

    if time_index:
        date_index = timestamps_serie.apply(handlers.time_helper.convert_milliseconds_to_time_zone_datetime, timezoned=time_zone)
        df.set_index(date_index, inplace=True)

    index_name = f"{symbol} {time_zone}"
    df.index.name = index_name

    if 'quoteQty' in columns.keys():
        return df[['Trade Id', 'Price', 'Quantity', 'Quote quantity', 'Date', 'Timestamp', 'Buyer was maker', 'Best price match']]
    else:  # it was a redis response
        return df[['Trade Id', 'Price', 'Quantity', 'Buyer Order Id', 'Seller Order Id', 'Date', 'Timestamp', 'Buyer was maker',
                   'Best price match']]


#############
# Orderbook #
#############


def get_order_book(symbol: str, limit=5000) -> dict:
    """
    Returns a dictionary with timestamp, bids and asks. Bids and asks are a list of lists with strings representing price and quantity.

    :param str symbol: A Binance valid symbol.
    :param int limit: Max is 5000. Default 5000.
    :return dict: A dict with ask and bids

    .. code-block::

        {
          "lastUpdateId": 1027024,
          "bids": [
            [
              "4.00000000",     // PRICE
              "431.00000000"    // QTY
            ]
          ],
          "asks": [
            [
              "4.00000200",
              "12.00000000"
            ]
          ]
        }

    """

    endpoint = '/api/v3/depth?'
    check_weight(limit // 100 or 1, endpoint=endpoint)

    query = {'symbol': symbol, 'limit': limit}
    return get_response(url=endpoint, params=query)


#####################
# Order Book Ticker #
#####################

def get_orderbook_tickers(symbol: str = None, decimal_mode: bool = False) -> dict:
    """
    Symbol Order Book Ticker


    GET /api/v3/ticker/bookTicker

    Best price/qty on the order book for a symbol or symbols.

    Weight(IP):

    Parameter	Symbols Provided	Weight
    symbol	                    1	    1
    symbol parameter is omitted	        2
    symbols	Any	                        2
    Parameters:

    Name	Type	Mandatory	Description
    symbol	STRING	NO	Parameter symbol and symbols cannot be used in combination.

    If neither parameter is sent, bookTickers for all symbols will be returned in an array.

    Examples of accepted format for the symbols parameter: ["BTCUSDT","BNBUSDT"]

    :param str symbol: Optional. If not passed, all symbols returned.
    :param bool decimal_mode: If selected, return decimal values.

    :return: Api response is:

        .. code-block::
            Response:

            {
              "symbol": "LTCBTC",
              "bidPrice": "4.00000000",
              "bidQty": "431.00000000",
              "askPrice": "4.00000200",
              "askQty": "9.00000000"
            }
            OR

            [
              {
                "symbol": "LTCBTC",
                "bidPrice": "4.00000000",
                "bidQty": "431.00000000",
                "askPrice": "4.00000200",
                "askQty": "9.00000000"
              },
              {
                "symbol": "ETHBTC",
                "bidPrice": "0.07946700",
                "bidQty": "9.00000000",
                "askPrice": "100000.00000000",
                "askQty": "1000.00000000"
              }
            ]

    """
    endpoint = '/api/v3/ticker/bookTicker?'
    if symbol:
        weight = 1
    else:
        weight = 2
    check_weight(weight, endpoint=endpoint)
    query = {'symbol': symbol}
    response = get_response(url=endpoint, params=query)
    if type(response) == dict:
        ret = {response['symbol']: response}
    else:
        ret = {i['symbol']: i for i in response}
    if decimal_mode:
        ret = {k: {kk: dd(vv) for kk, vv in v.items() if kk != 'symbol'} for k, v in ret.items()}
    else:
        ret = {k: {kk: float(vv) for kk, vv in v.items() if kk != 'symbol'} for k, v in ret.items()}
    return ret

####################
# coin conversions #
####################


def intermediate_conversion(coin: str,
                            decimal_mode: bool,
                            prices: dict = None,
                            try_coin: str = 'BTC',
                            coin_qty: float = 1) -> float or None:
    """
    Uses an intermediate symbol for conversion. Uses stablecoin USDT versus other "try" coin.

    :param str coin: A binance coin.
    :param bool decimal_mode: It flags to work in decimal mode.
    :param dict prices: BinPan prices dict.
    :param str try_coin: Coin to try as intermediate in case of not existing pair. Default is BTC.
    :param float coin_qty: Quantity to convert. Default is 1.
    :return float: converted value.
    """

    if not prices:
        prices = get_prices_dic(decimal_mode=decimal_mode)

    if coin + try_coin in prices.keys():
        price = prices[coin + try_coin]
        try_symbol = f"{try_coin}USDT"
        return price * coin_qty * prices[try_symbol]

    elif try_coin + coin in prices.keys():
        price = 1 / prices[try_coin + coin]
        try_symbol = f"USDT{try_coin}"
        try:
            return price * coin_qty * prices[try_symbol]
        except KeyError:
            return None
    else:
        return None


def convert_coin(coin: str,
                 decimal_mode: bool,
                 convert_to: str = 'BUSD',
                 coin_qty: float or dd = 1,
                 prices: dict = None) -> float or None:
    """
    Calculates a coin quantity value converted to other coin with current exchange prices.

    :param str coin: An existing coin string.
    :param bool decimal_mode: It flags to work in decimal mode.
    :param str convert_to: An existing coin string.
    :param float coin_qty: How many coins to convert to.
    :param dict prices: A dictionary with symbols and prices.
    :return float: Converted value for the quantity
    """
    coin = coin.upper()
    convert_to = convert_to.upper()

    if coin == convert_to:
        return coin_qty

    if not prices:
        prices = get_prices_dic(decimal_mode=decimal_mode)

    symbol_a = coin + convert_to
    symbol_b = convert_to + coin

    if symbol_a in prices.keys():
        return coin_qty * prices[symbol_a]

    elif symbol_b in prices.keys():
        return coin_qty * (1 / prices[symbol_b])
    else:
        ret1 = intermediate_conversion(coin=coin, prices=prices, try_coin='BTC', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret1:
            return ret1
        ret2 = intermediate_conversion(coin=coin, prices=prices, try_coin='BUSD', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret2:
            return ret2
        ret3 = intermediate_conversion(coin=coin, prices=prices, try_coin='BNB', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret3:
            return ret3
        ret4 = intermediate_conversion(coin=coin, prices=prices, try_coin='ETH', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret4:
            return ret4
        ret5 = intermediate_conversion(coin=coin, prices=prices, try_coin='TUSD', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret5:
            return ret5
        ret6 = intermediate_conversion(coin=coin, prices=prices, try_coin='USDC', coin_qty=coin_qty, decimal_mode=decimal_mode)
        if ret6:
            return ret6
        else:
            market_logger.warning(f"No possible conversion for {coin} to {convert_to}")
            return None



from setuptools import setup
import pathlib

try:
    from secret import version
except ImportError:
    version = "0.2.45"

# import configparser

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

readme = open('README.md', 'r')
README_TEXT = readme.read()
readme.close()

read_required = open('requirements.txt', 'r')
REQUIRED = read_required.read()
read_required.close()

my_version = version

setup(name='binpan',
      version=my_version,
      url='https://github.com/nand0san/binpan_studio',
      license='MIT',
      install_requires=REQUIRED,
      python_requires='>=3.7.9',
      classifiers=[
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.7",
      ],
      author='Fernando Alfonso',
      author_email='hancaidolosdos@hotmail.com',
      description='Binance API wrapper with backtesting tools.',
      long_description=README_TEXT,
      long_description_content_type="text/markdown",
      package_dir={"": "."},
      packages=["handlers", "binpan"]
      )



import requests
from base64 import b64encode, b64decode
import hashlib
from binascii import unhexlify
from os.path import expanduser
from cpuinfo import get_cpu_info
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad


class AesCipher(object):

    def __init__(self):
        __seed = bytes(expanduser("~") + get_cpu_info()['brand_raw'], "utf8")
        self.__iv = hashlib.md5(__seed).hexdigest()
        self.__key = hashlib.md5(__seed[::-1]).hexdigest()

    def encrypt(self, msg):
        msg_padded = pad(msg.encode(), AES.block_size)
        cipher = AES.new(unhexlify(self.__key), AES.MODE_CBC, unhexlify(self.__iv))
        cipher_text = cipher.encrypt(msg_padded)
        return b64encode(cipher_text).decode('utf-8')

    def decrypt(self, msg_encrypted):
        decipher = AES.new(unhexlify(self.__key), AES.MODE_CBC, unhexlify(self.__iv))
        plaintext = unpad(decipher.decrypt(b64decode(msg_encrypted)), AES.block_size).decode('utf-8')
        return plaintext


def get_exchange_limits() -> dict:
    """
    Binance manage several limits: RAW_REQUESTS, REQUEST_WEIGHT, and ORDERS rate limits.

    The headers for those limits, I assume that are:
    - RAW_REQUESTS: x-mbx-used-weight. Is cross all the api calls.
    - REQUEST_WEIGHT: Example: x-mbx-order-count-10s. Time interval limited requests.
    - ORDERS: Example: x-mbx-order-count-10s. Rate limit for orders.
    - X-SAPI-USED-IP-WEIGHT-1M: For sapi endpoint requests.

    Example response:

        {'X-SAPI-USED-IP-WEIGHT-1M': 1200,
         'x-mbx-order-count-10s': 50,
         'x-mbx-order-count-1d': 160000,
         'x-mbx-used-weight': 6100,
         'x-mbx-used-weight-1m': 1200}

    :return dict:
    """
    base_url = 'https://api.binance.com'
    endpoint = '/api/v3/exchangeInfo'
    response = requests.get(base_url+endpoint).json()
    # info_dic = {k['symbol']: k for k in response['symbols']}

    limits = response['rateLimits']
    limits_dict = {}
    for limit in limits:
        if 'REQUEST' in limit['rateLimitType']:
            interval = str(limit['intervalNum'])
            interval += limit['interval'][0]
            limits_dict[f'REQUEST_{interval}'] = limit['limit']
        elif 'ORDERS' in limit['rateLimitType']:
            interval = str(limit['intervalNum'])
            interval += limit['interval'][0]
            limits_dict[f'ORDERS_{interval}'] = limit['limit']
        else:
            msg = f"BinPan error: Unknown limit from API: {limit}"
            raise Exception(msg)
    #
    #
    #
    # for i in limits:
    #     if i['rateLimitType'].upper() == 'ORDERS':
    #         k1 = f"x-mbx-order-count-{i['intervalNum']}{i['interval'][0].lower()}"
    #         k2 = f"x-mbx-order-count-{i['intervalNum']}{i['interval'][0].lower()}"
    #
    #     elif i['rateLimitType'].upper() == 'REQUEST_WEIGHT':
    #         k1 = f"x-mbx-used-weight-{i['intervalNum']}{i['interval'][0].lower()}"
    #         k2 = f"X-SAPI-USED-IP-WEIGHT-{i['intervalNum']}{i['interval'][0].upper()}"
    #
    #     elif i['rateLimitType'].upper() == 'RAW_REQUESTS':
    #         k1 = "x-mbx-used-weight"
    #         k2 = "x-mbx-used-weight"
    #     else:
    #         raise Exception("BinPan Rate Limit not parsed")
    #
    #     v = i['limit']
    #
    #     limits_dict[k1] = v
    #     limits_dict[k2] = v

    return limits_dict
